import core_types
import ctypes
import numpy as np
import type_conv

from core_types import StructT, IncompatibleTypes, Int64, ptr_type
from tuple_type import TupleT, repeat_tuple
from core_types import TypeValueT

def buffer_info(buf, ptr_type = ctypes.c_void_p):
  """Given a python buffer, return its address and length"""

  assert isinstance(buf, buffer)
  address = ptr_type()
  length = ctypes.c_ssize_t()
  obj = ctypes.py_object(buf)
  ctypes.pythonapi.PyObject_AsReadBuffer(obj, ctypes.byref(address),
                                         ctypes.byref(length))
  return address, length.value

ctypes.pythonapi.PyBuffer_New.argtypes = (ctypes.c_ulong,)
ctypes.pythonapi.PyBuffer_New.restype = ctypes.py_object

PyBuffer_New = ctypes.pythonapi.PyBuffer_New
PyBuffer_FromReadWriteMemory = ctypes.pythonapi.PyBuffer_FromReadWriteMemory

class SliceT(StructT):
  _members = ['start_type', 'stop_type', 'step_type']

  def node_init(self):
    self._fields_ = [
      ('start', self.start_type),
      ('stop', self.stop_type),
      ('step', self.step_type),
    ]

  def __eq__(self, other):
    return self is other or \
      (other.__class__ is SliceT and
       self.start_type == other.start_type and
       self.stop_type == other.stop_type and
       self.step_type == other.step_type)

  def __hash__(self):
    return hash((self.start_type, self.stop_type, self.step_type))

  def combine(self, other):
    if self == other:
      return self
    else:
      raise IncompatibleTypes(self, other)

  def __str__(self):
    return "SliceT(%s, %s, %s)" % (self.start_type, self.stop_type,
                                   self.step_type)

  def __repr__(self):
    return str(self)

  def from_python(self, py_slice):
    start = self.start_type.from_python(py_slice.start)
    stop = self.stop_type.from_python(py_slice.stop)
    step = self.step_type.from_python(py_slice.step)
    return self.ctypes_repr(start, stop, step)

  def to_python(self, obj):
    start = self.start_type.to_python(obj.start)
    stop = self.stop_type.to_python(obj.stop)
    step = self.step_type.to_python(obj.step)
    return slice(start, stop, step)

_slice_type_cache = {}
def make_slice_type(start_t, stop_t, step_t):
  key = (start_t, stop_t, step_t)
  if key in _slice_type_cache:
    return _slice_type_cache[key]
  else:
    t = SliceT(start_t, stop_t, step_t)
    _slice_type_cache[key] = t
    return t

def typeof_slice(s):
  start_type = type_conv.typeof(s.start)
  stop_type = type_conv.typeof(s.stop)
  step_type = type_conv.typeof(s.step)
  return make_slice_type(start_type, stop_type, step_type)

type_conv.register(slice, SliceT, typeof_slice)

class ArrayT(StructT):
  _members = ['elt_type', 'rank']

  def node_init(self):
    
    assert isinstance(self.elt_type, core_types.ScalarT), \
      "Can't create array with element type %s, currently only scalar elements supported" % \
      (self.elt_type,)
        
    tuple_t = repeat_tuple(Int64, self.rank)

    self.shape_t = tuple_t
    self.strides_t = tuple_t
    self.ptr_t = ptr_type(self.elt_type)
    self._fields_ = [
      ('data', self.ptr_t),
      ('shape', tuple_t),
      ('strides', tuple_t),
      ('offset', Int64),
      ('size', Int64),
      # ('dtype', TypeValueT(self.elt_type))
    ]
  
  def dtype(self):
    return self.elt_type.dtype

  def __str__(self):
    return "array%d(%s)" % (self.rank, self.elt_type)

  def __eq__(self, other):
    return other.__class__ is ArrayT and \
        self.elt_type == other.elt_type and \
        self.rank == other.rank

  def __hash__(self):
    return hash((self.elt_type, self.rank))

  def combine(self, other):
    if self == other:
      return self
    elif isinstance(other, core_types.ScalarT):
      elt_t = self.elt_type
      combined_elt_t = elt_t.combine(other)
      return make_array_type(combined_elt_t, self.rank)
    elif isinstance(other, ArrayT):
      assert self.rank == other.rank
      combined_elt_t = self.elt_type.combine(other.elt_type)
      return make_array_type(combined_elt_t, self.rank)
    else:
      raise IncompatibleTypes(self, other)

  def index_type(self, idx):
    """
    Given the type of my indices, what type of array will result?
    """

    if not isinstance(idx, core_types.Type):
      idx = idx.type

    if isinstance(idx, TupleT):
      indices = idx.elt_types
    else:
      indices = [idx]

    n_indices = len(indices)
    n_required = self.rank
    if n_required > n_indices:
      n_missing = n_required - n_indices
      extra_indices = [core_types.NoneType] * n_missing
      indices = indices + extra_indices

    # we lose one result dimension for each int in the index set
    result_rank = n_required
    for t in indices:
      if isinstance(t, core_types.IntT):
        result_rank -= 1
      else:
        assert isinstance(t, (core_types.NoneT, SliceT, ArrayT, TupleT)), \
            "Unexpected index type: %s " % t
    if result_rank > 0:
      return make_array_type(self.elt_type, result_rank)
    else:
      return self.elt_type

  # WARNING:
  # until we have garbage collection figured out, we'll
  # leak memory from arrays we allocate in the conversion routine
  _store_forever = []
  _seen_ptr = set([])
  def from_python(self, x):
    if not isinstance(x, np.ndarray):
      print "Warning: Copying %s into Parakeet, will never get deleted" % \
          (type(x))
      x = np.asarray(x)
      self._store_forever.append(x)

    
    nelts = reduce(lambda x,y: x*y, x.shape)
    elt_size = x.dtype.itemsize
    # total_bytes = nelts * elt_size
    if x.base is not None:
      if isinstance(x.base, np.ndarray):
        ptr = x.base.ctypes.data_as(self.ptr_t.ctypes_repr)
        offset_bytes = x.ctypes.data  - x.base.ctypes.data 
      else:
        assert isinstance(x.base, buffer)
        ptr, _ = buffer_info(x.base, self.ptr_t.ctypes_repr)
        offset_bytes = x.ctypes.data - ctypes.addressof(ptr.contents) 
      offset = offset_bytes / elt_size
    else:
      ptr = x.ctypes.data_as(self.ptr_t.ctypes_repr)
      offset = 0
   
    ctypes_shape = self.shape_t.from_python(x.shape)
    strides_in_elts = tuple([s / elt_size for s in x.strides])
    ctypes_strides = self.strides_t.from_python(strides_in_elts)
    return self.ctypes_repr(ptr, 
                            ctypes.pointer(ctypes_shape),
                            ctypes.pointer(ctypes_strides), 
                            offset, 
                            nelts)
    
  def to_python(self, obj):
    """
    For now, to avoid to dealing with the messiness of ownership, we just always
    copy data on the way out of Parakeet
    """

    shape = self.shape_t.to_python(obj.shape.contents)

    elt_size = self.elt_type.nbytes
    strides_in_elts = self.strides_t.to_python(obj.strides.contents)
    strides_in_bytes = tuple([s * elt_size for s in strides_in_elts])

    base_ptr = obj.data

    nbytes = obj.size * elt_size

    dest_buf = PyBuffer_New(nbytes)
    dest_ptr, _ = buffer_info(dest_buf, self.ptr_t.ctypes_repr)
    ctypes.memmove(dest_ptr, base_ptr, nbytes)

    
    return np.ndarray(shape, dtype = self.elt_type.dtype,
                      buffer = dest_buf,
                      strides = strides_in_bytes,
                      offset = obj.offset * elt_size)

_array_types = {}
def make_array_type(elt_t, rank):
  key = (elt_t, rank)
  if key in _array_types:
    return _array_types[key]
  else:
    if rank == 0:
      t = elt_t
    else:
      t = ArrayT(elt_t, rank)
    _array_types[key] = t
    return t


def elt_type(t):
  if isinstance(t, ArrayT):
    return t.elt_type
  else:
    return t

def elt_types(ts):
  return map(elt_type, ts)

def lower_rank(t, r):
  if isinstance(t, ArrayT):
    assert t.rank >= r
    return make_array_type(t.elt_type, t.rank - r)
  else:
    return t

def increase_rank(t, r):
  if isinstance(t, ArrayT):
    return make_array_type(t.elt_type, t.rank + r)
  else:
    return make_array_type(t, r)

def lower_ranks(arg_types, r):
  return [lower_rank(t, r) for t in arg_types]

def get_rank(t):
  if isinstance(t, ArrayT):
    return t.rank
  else:
    return 0

def rank(t):
  return get_rank(t)




import names  
import core_types
import syntax
import syntax_helpers
import tuple_type

from array_type import ArrayT
from core_types import Int64 
from tuple_type import TupleT 
from syntax import Assign, Tuple, TupleProj, Var, Cast, Return, Index  
from syntax_helpers import get_types, zero_i64, one_i64, none 
from transform import Transform

class RewriteTyped(Transform):
  def __init__(self, return_type = None):
    Transform.__init__(self, verify = False)
    self.forced_return_type = return_type 
    
    
  def pre_apply(self, fn):
    
    if self.forced_return_type is None:
      self.fn_return_type = self.fn.type_env["$return"]
    else:
      self.fn_return_type = self.forced_return_type
      
  
  def post_apply(self, fn):
    fn.return_type = self.fn_return_type  
  
  def coerce_expr(self, expr, t):
    assert t is not None
    expr = self.transform_expr(expr)
    if expr.type == t:
      return expr
    elif expr.__class__ is Tuple:
      if t.__class__ is not TupleT or \
          len(expr.type.elt_types) != t.elt_types:
        raise core_types.IncompatibleTypes(expr.type, t)
      else:
        new_elts = []
        for elt, elt_t in zip(expr.elts, t.elt_types):
          new_elts.append(self.coerce_expr(elt, elt_t))
        return Tuple(new_elts, type = t)
    else:
      assert \
          isinstance(expr.type, core_types.ScalarT) and \
          isinstance(t, core_types.ScalarT), \
          "Can't cast type %s into %s" % (expr.type, t)
      return syntax.Cast(expr, type=t)

  def transform_merge(self, merge):
    new_merge = {}
    for (var, (left, right)) in merge.iteritems():
      t = self.type_env[var]
      new_left = self.coerce_expr(left, t)
      new_right = self.coerce_expr(right, t)
      new_merge[var] = (new_left, new_right)
    return new_merge

  def transform_lhs(self, lhs):
    if isinstance(lhs, Var):
      t = self.lookup_type(lhs.name)
      if t == lhs.type:
        return lhs
      else:
        return Var(lhs.name, type = t)
    elif isinstance(lhs, syntax.Tuple):
      elts = map(self.transform_lhs, lhs.elts)
      elt_types = get_types(elts)
      if elt_types != lhs.type.elt_types:
        return syntax.Tuple(elts, type = tuple_type.make_tuple_type(elt_types))
      else:
        return lhs
    else:
      return lhs

  def transform_Var(self, expr):
    expr.type = self.fn.type_env[expr.name]
    return expr

  def transform_PrimCall(self, expr):
      arg_types = get_types(expr.args)
      upcast_types = expr.prim.expected_input_types(arg_types)
      result_type = expr.prim.result_type(upcast_types)
      upcast_args = [self.coerce_expr(x, t)
                     for (x,t) in zip(expr.args, upcast_types)]
      return syntax.PrimCall(expr.prim, upcast_args, type = result_type)

  def transform_Array(self, expr):
    array_t = expr.type
    elt_t = array_t.elt_type
    assert array_t.rank > 0
    if array_t.rank == 1:
      new_elts = [self.coerce_expr(elt, elt_t) for elt in expr.elts]
      return syntax.Array(new_elts, type = array_t)
    else:
      # need to allocate an output array and copy the elements in
      first_elt = self.assign_temp(expr.elts[0], "first_elt")
      elt_dims = [self.shape(first_elt, i) for i in xrange(array_t.rank - 1)]
      n = len(expr.elts)
      outer_dim = syntax_helpers.const(n)
      all_dims = (outer_dim,) + tuple(elt_dims)
      array = self.alloc_array(elt_t, all_dims, "array_literal")
      for i, elt in enumerate(expr.elts):
        idx_expr = self.index(array, i, temp = False)
        # transform indexing to make missing indices explicit
        self.assign(idx_expr, expr.elts[i])
      return array

  def transform_Reduce(self, expr):
    acc_type = self.return_type(expr.combine)
    if expr.init and \
        not self.is_none(expr.init) and \
        expr.init.type != acc_type:
      assert len(expr.args) == 1
      expr.init = self.coerce_expr(expr.init, acc_type)
    return expr


    
  def transform_Scan(self, expr):
    acc_type = self.return_type(expr.combine)
    if expr.init and \
        not self.is_none(expr.init) and \
        expr.init.type != acc_type:
      expr.init = self.coerce_expr(expr.init, acc_type)
    return expr

  def transform_Slice(self, expr):
    """
    # None step defaults to 1
    if isinstance(expr.step.type, core_types.NoneT):
      start_t = expr.start.type
      stop_t = expr.stop.type
      step = syntax_helpers.one_i64
      step_t = step.type
      slice_t = array_type.make_slice_type(start_t, stop_t, step_t)
      expr.step = step
      expr.type = slice_t
    """
    return expr
  
  index_function_cache = {}
  def get_index_fn(self, array_t, idx_t):
    key = (array_t, idx_t) 
    if key in self.index_function_cache:
      return self.index_function_cache[key]
    array_name = names.fresh("array")
    array_var = Var(array_name, type = array_t)
    idx_name = names.fresh("idx")
    idx_var = Var(idx_name, type = idx_t)
    if idx_t is not Int64:
      idx_var = Cast(value = idx_var,  type = Int64)
    elt_t = array_t.index_type(idx_t)

    fn = syntax.TypedFn(
        name = names.fresh("idx"), 
        arg_names = (array_name, idx_name),
        input_types = (array_t, idx_t),
        return_type = elt_t,
        type_env = {array_name:array_t, idx_name:idx_t}, 
        body = [Return (Index(array_var, idx_var, type = elt_t))]) 
    self.index_function_cache[key] = fn 
    return fn 
    
  def transform_Index(self, expr):
    # TODO: Make fancy indexing work 
    # with multiple indices, boolean index elements, 
    # and multi-dimensional indexing
    
    index = expr.index
    #if index.type.__class__ is TupleT:
    #  indices = index.elts 
    #else:
    #  indices = [index]
    
    if index.type.__class__ is ArrayT:
      assert index.type.rank == 1, \
        "Don't yet support indexing by %s" % index.type 
      
      index_elt_t = index.type.elt_type
      assert False
      #if index_elt_t == core_types.Bool:
      #  index_array = syntax.Where(expr.index)
      #  index_elt_t = core_types.Int64
      #else:
      #  index_array = expr.index 
      index_array = expr.index 
      index_fn = self.get_index_fn(expr.value.type, index_elt_t)
      index_closure = self.closure(index_fn, [expr.value])
      return syntax.Map(fn = index_closure, 
                        args = index_array, 
                        type = expr.value.type, 
                        axis = zero_i64)
    else:
      return expr 
  
  def transform_Assign(self, stmt):
    new_lhs = self.transform_lhs(stmt.lhs)
    lhs_t = new_lhs.type
    rhs = self.transform_expr(stmt.rhs)
    assert lhs_t is not None, "Expected a type for %s!" % stmt.lhs
    if new_lhs.__class__ is Tuple: 
      rhs = self.assign_temp(rhs)
      for (i, lhs_elt) in enumerate(new_lhs.elts):
        if lhs_elt.__class__ is Var:
          name = names.original(lhs_elt.name) 
        else:
          name = None 
        idx = self.index(rhs, i, name = name )
        elt_stmt = self.transform_Assign(Assign(lhs_elt, idx))
        self.blocks.append_to_current(elt_stmt)
      return None  
    else:         
      
      new_rhs = self.coerce_expr(rhs, lhs_t)
      assert new_rhs.type and isinstance(new_rhs.type, core_types.Type), \
          "Expected type annotation on %s, but got %s" % (new_rhs, new_rhs.type)
      stmt.lhs = new_lhs
      stmt.rhs = new_rhs
      return stmt

  def transform_If(self, stmt):
    stmt.cond = self.coerce_expr(stmt.cond, core_types.Bool)
    stmt.true = self.transform_block(stmt.true)
    stmt.false = self.transform_block(stmt.false)
    stmt.merge = self.transform_merge(stmt.merge)
    return stmt

  def transform_Return(self, stmt):
    stmt.value = self.coerce_expr(stmt.value, self.fn_return_type)
    return stmt

  def transform_While(self, stmt):
    stmt.cond = self.coerce_expr(stmt.cond, core_types.Bool)
    stmt.body = self.transform_block(stmt.body)
    stmt.merge = self.transform_merge(stmt.merge)
    return stmt

def rewrite_typed(typed_fundef, return_type = None):
  return RewriteTyped(return_type).apply(typed_fundef)



import numpy as np
import scipy.spatial
import time

import parakeet
from parakeet import allpairs, each
from testing_helpers import eq, run_local_tests

def python_update_assignments(X, centroids):
  dists = scipy.spatial.distance.cdist(X, centroids, 'sqeuclidean')
  return np.argmin(dists, 1)

def python_update_centroids(X, assignments, k):
  d = X.shape[1]
  new_centroids = np.zeros((k,d), dtype=X.dtype)
  for i in xrange(k):
    mask = assignments == i
    count = np.sum(mask)
    assigned_data = X[mask]
    if count > 1:
      new_centroids[i, :] = np.mean(assigned_data)
    elif count == 1:
      new_centroids[i,:] = assigned_data[0]

  return new_centroids

def python_kmeans(X, k, maxiters = 100, initial_assignments = None):
  n = X.shape[0]
  if initial_assignments is None:
    assignments = np.random.randint(0, k, size=n)
  else:
    assignments = initial_assignments
  centroids = python_update_centroids(X, assignments, k)
  for _ in xrange(maxiters):
    old_assignments = assignments

    assignments = python_update_assignments(X, centroids)

    if all(old_assignments == assignments):
      break
    centroids = python_update_centroids(X, assignments, k)

  return centroids

def sqr_dist(x,y):
  return sum((x-y) ** 2)

def parakeet_update_assignments(X, centroids):
  dists = allpairs(sqr_dist, X, centroids)

  return np.argmin(dists, 1)

def mean(X):
  return sum(X) / len(X)

def parakeet_update_centroids(X, assignments, k):
  d = X.shape[1]

  new_centroids = np.zeros((k,d), dtype=X.dtype)
  for i in xrange(k):
    mask = (assignments == i)
    count = np.sum(mask)
    assigned_data = X[mask]
    if count == 1:
      new_centroids[i, :] = assigned_data[0]
    elif count > 1:
      new_centroids[i,:] = parakeet.mean(assigned_data)
  return new_centroids

def parakeet_kmeans(X, k, maxiters = 100, initial_assignments = None):
  n = X.shape[0]
  if initial_assignments is None:
    assignments = np.random.randint(0, k, size = n)
  else:
    assignments = initial_assignments

  centroids = python_update_centroids(X, assignments, k)
  for _ in xrange(maxiters):
    old_assignments = assignments
    assignments = parakeet_update_assignments(X, centroids)
    if all(old_assignments == assignments):
      break
    centroids = python_update_centroids(X, assignments, k)

  return centroids

def test_kmeans():
  n = 200
  d = 4
  X = np.random.randn(n*d).reshape(n,d)
  k = 2
  niters = 10
  assignments = np.random.randint(0, k, size = n)
  parakeet_C = parakeet_kmeans(X, k, niters, assignments)
  python_C = python_kmeans(X, k, niters, assignments)
  assert eq(parakeet_C.shape, python_C.shape), \
      "Got back wrong shape, expect %s but got %s" % \
      (python_C.shape, parakeet_C.shape)
  assert eq(parakeet_C, python_C), \
      "Expected %s but got %s" % (python_C, parakeet_C)

def test_kmeans_perf():
  n = 160
  d = 20
  X = np.random.randn(n*d).reshape(n,d)
  k = 5 
  niters = 2
  assignments = np.random.randint(0, k, size = n)
#
#  start = time.time()
#  _ = python_kmeans(X, k, niters, assignments)
#  python_time = time.time() - start

  # run parakeet once to warm up the compiler
  start = time.time()
  _ = parakeet_kmeans(X, k, niters, assignments)
  parakeet_with_comp = time.time() - start

  start = time.time()
  _ = parakeet_kmeans(X, k, niters, assignments)
  parakeet_time = time.time() - start

  #speedup = python_time / parakeet_time
  print "Parakeet time:", parakeet_with_comp
  print "Parakeet w/out compilation:", parakeet_time
  #print "Python time", python_time

#  assert speedup > 1, \
#      "Parakeet too slow! Python time = %s, Parakeet = %s, %.1fX slowdown " % \
#      (python_time, parakeet_time, 1/speedup)

if __name__ == '__main__':
  run_local_tests()

"""
parakeet.config.call_from_python_in_parallel = False

py_reduce = reduce
py_sum = sum

from parakeet import  sum, reduce, each
from prims import add
from testing_helpers import eq, run_local_tests

def ident(x):
  return x

def sqr_dist(x, y):
  return sum((x-y)*(x-y))

def argmin((curmin, curminidx), (val, idx)):
  if val < curmin:
    return (val, idx)
  else:
    return (curmin, curminidx)

inf = np.inf
def py_minidx(C, idxs, x):
  def run_sqr_dist(c):
    return sqr_dist(c, x)
  dists = map(run_sqr_dist, C)
  zipped = zip(dists, idxs)
  return py_reduce(argmin, zipped, (inf,-1))[1]

def py_calc_centroid(X, a, i):
  return np.mean(X[a == i], axis=0)

def calc_centroid(X, a, i):
  subset = X[a==i]
  return reduce(add, subset) / subset.shape[0]

def parakeet_zip(xs, ys):
  def make_pair(x, y):
    return (x, y)
  return each(make_pair, xs, ys)

def par_minidx(C, idxs, x):
  def run_sqr_dist(c):
    return sqr_dist(c, x)
  dists = each(run_sqr_dist, C)
  zipped = parakeet_zip(dists, idxs)
  return reduce(argmin, zipped, init=(inf,-1))[1]

def py_dists(C, x):
  def run_sqr_dist(c):
    return sqr_dist(c, x)
  return np.array(map(run_sqr_dist, C))

def par_dists(C, x):
  def run_sqr_dist(c):
    return sqr_dist(c, x)
  return each(run_sqr_dist, C)

def par_kmeans(X, assign, k, max_iters = 100):
  centroid_indices = np.arange(k)
  def run_calc_centroid(i):
    return calc_centroid(X, assign, i)
  C = parakeet.each(run_calc_centroid,
                    centroid_indices)
  for iter_num in xrange(max_iters):
    lastAssign = assign
    def run_minidx(i):
      return par_minidx(C, centroid_indices, i)
    assign = parakeet.each(run_minidx, X)
    if np.all(assign == lastAssign):
      break
    C = parakeet.each(run_calc_centroid,
                      centroid_indices)
  return C, assign

def kmeans(X, assign, k):
  idxs = np.arange(k)
  def run_calc_centroid(i):
    return py_calc_centroid(X, assign, i)

  C = np.array(map(run_calc_centroid, idxs))
  converged = False
  j = 0
  while not converged and j < 100:
    lastAssign = assign
    def run_py_minidx(x):
      return py_minidx(C, idxs, x)

    assign = np.array(map(run_py_minidx, X))
    converged = np.all(assign == lastAssign)
    C = np.array(map(run_calc_centroid, idxs))
    j = j + 1
  return C, assign
"""



import config 

from core_types import ScalarT 
from syntax import Var, Attribute, Tuple 
from syntax_visitor import SyntaxVisitor

empty = set([])

def collect_nonscalar_names(expr):
  if expr is None or isinstance(expr.type, ScalarT):
    return []
  elif expr.__class__ is Var:
    return [expr.name]
  else:
    return collect_nonscalar_names_from_list(expr.children())

def collect_nonscalar_names_from_list(exprs):
  result = []
  for expr in exprs:
    result.extend(collect_nonscalar_names(expr))
  return result

class EscapeAnalysis(SyntaxVisitor):
  """
  A very imprecise combined escape and alias analysis. 
  Rough idea: whenever you assign some 
     x = expr(y,z)
  then take all the non-scalar values and unify them into 
  a single alias set along with all the previous variable
  names in the alias sets of y and z.  
  """
  
  def visit_fn(self, fn):
    self.scalars = set([])
    self.may_alias = {}
    all_scalars = True 
    # every name at least aliases it selfcollect_var_names
    for (name,t) in fn.type_env.iteritems():
      if isinstance(t, ScalarT):
        self.scalars.add(name)
      else:
        self.may_alias[name] = set([name])
        all_scalars = False 
    
    self.may_escape = set([])
    
    if all_scalars:
      return  
    
    self.visit_block(fn.body)
    
    # once we've accumulated all the aliases
    # anyone who points into the input data 
    # is also considered to be escaping 

    for name in fn.arg_names:
      if name not in self.scalars:
        self.may_escape.update(self.may_alias[name])
    
    if config.print_escape_analysis: 
      print "[EscapeAnalysis] In function %s (version %d)" % \
         (fn.name, fn.version) 
      print "-------"
      print fn 
      print 
      print "aliases"
      print "-------"
      for (k,aliases) in sorted(self.may_alias.items(), key = lambda (k,_): k):
        print "  %s => %s" % (k, aliases)
      print 
      print "escape set"
      print "----------"
      print sorted(self.may_escape)

  def mark_escape(self, name):
    if name not in self.scalars:
      for alias in self.may_alias[name]:
        self.may_escape.add(alias)

  def mark_escape_list(self, names):
    for name in names:
        self.mark_escape(name)
  
  def update_aliases(self, lhs_name, rhs_names):
    """
    Once we've gotten all the RHS names being assigned to the 
    LHS var, we do a recursively lookup into may_alias since
    we may have code like:
      a = tuple(b,c)
      d = tuple(a,e) 
    and we want the alias set of d to be {a,e,b,c}
    """
    if lhs_name not in self.scalars:
      combined_set = self.may_alias[lhs_name]
      for rhs_name in rhs_names:
        if rhs_name not in self.scalars:
          combined_set.update(self.may_alias[rhs_name])
      for alias in combined_set:
        self.may_alias[alias] = combined_set
      return combined_set  
  
  def update_escaped(self, lhs_name, rhs_alias_set):
    if lhs_name not in self.scalars and \
       any(alias in self.may_escape for alias in rhs_alias_set):
        self.may_escape.update(rhs_alias_set.difference(self.scalars))
        self.may_escape.add(lhs_name)
  
  def visit_Call(self, expr):
    self.mark_escape_list(collect_nonscalar_names_from_list(expr.args))

  def collect_lhs_names(self, expr):
    if expr.__class__ is Var:
      return [expr.name]
    elif expr.__class__ is Attribute:
      return self.collect_lhs_names(expr.value)
    elif expr.__class__ is Tuple:
      combined = []
      for elt in expr.elts:
        combined.extend(self.collect_lhs_names(elt))
    else:
      return []

  def visit_Assign(self, stmt):
    lhs_names = set(self.collect_lhs_names(stmt.lhs))
    rhs_names = set(collect_nonscalar_names(stmt.rhs))
    for lhs_name in lhs_names:
      self.update_aliases(lhs_name, rhs_names)

  def visit_Return(self, expr):
    self.mark_escape_list(collect_nonscalar_names(expr.value))

  def visit_merge(self, merge):
    for (name, (l,r)) in merge.iteritems():
      if l.__class__ is Var: 
        left_aliases = self.may_alias.get(l.name, empty)
      else:
        left_aliases = empty 
      if r.__class__ is Var: 
        right_aliases = self.may_alias.get(r.name, empty)
      else:
        right_aliases = empty
      combined_set = self.update_aliases(name, left_aliases.union(right_aliases))
      self.update_escaped(name, combined_set)

_cache = {}
def run(fundef):
  key = (fundef.name, fundef.copied_by, fundef.version)
  if key in _cache:
    return _cache[key]
  else: 
    analysis = EscapeAnalysis()
    analysis.visit_fn(fundef)
    _cache[key] = analysis
    return analysis

def may_alias(fundef):
  return run(fundef).may_alias 

def may_escape(fundef):
  return run(fundef).may_escape



import numpy as np 
import parakeet 

import testing_helpers 

values = [np.array(0), np.array(0.0), np.array(True), 
          np.array([0]), 
          np.array([[0],[1]]), 
          np.array([[0.0], [1.0]]), 
          np.array([[True], [False]])
         ]

def run(fn, prop ):
  fn = parakeet.jit(fn)
  for v in values:
    expected = getattr(v, prop)
    result = fn(v)
    assert np.allclose(expected, result), "Expected %s but got %s" % (expected, result)
  

def test_prop_real():
  def get_real(x):
    return x.real 
  run(get_real, 'real')

#def test_prop_imag():
#  def get_imag(x):
#    return x.imag 
#  run(get_imag, 'imag')

def test_prop_ndim():
  def get_ndim(x):
    return x.ndim 
  run(get_ndim, 'ndim')

def test_prop_size():
  def get_size(x):
    return x.size
  run(get_size, 'size')

def test_prop_shape():
  def get_shape(x):
    return x.shape 
  run(get_shape, 'shape')

def test_prop_T():
  def get_T(x):
    return x.T
  run(get_T, 'T')

if __name__ == '__main__':
  testing_helpers.run_local_tests()



class ScopedDictionary(object):
  def __init__(self):
    self.scopes = []
    
  def push(self, d = None):
    if d is None:
      d = {}
    self.scopes.append(d)
  
  def pop(self):
    return self.scopes.pop()
  
  def top(self):
    return self.scopes[-1]
  
  def get(self, key):
    for scope in reversed(self.scopes):
      res = scope.get(key)
      if res:
        return res 
    return None
  
  def __getitem__(self, key):
    for scope in reversed(self.scopes):
      res = scope.get(key)
      if res:
        return res 
    assert False, "Key %s not found" % key
    
  def __setitem__(self, key, value):
    self.scopes[-1][key] = value 
    
  def setdefault(self, key, value):
    self.scopes[-1].setdefault(key, value)
  
  def __contains__(self, key):
    for scope in reversed(self.scopes):
      if key in scope:
        return True 
    return False 


import itertools 
import numpy as np
import types

import adverb_semantics
import ast_conversion
import syntax

from args import ActualArgs
from common import dispatch
from core_types import ScalarT, StructT

class InterpSemantics(adverb_semantics.AdverbSemantics):
  def size_along_axis(self, value, axis):
    assert len(value.shape) > axis, \
        "Can't get %d'th element of %s with shape %s" % \
        (axis, value, value.shape)
    return value.shape[axis]

  def is_tuple(self, x):
    return isinstance(x, tuple)

  def is_none(self, x):
    return x is None

  def rank(self, value):
    return np.rank(value)

  def int(self, x):
    return int(x)

  def bool(self, x):
    return bool(x)

  def add(self, x, y):
    return x + y

  def sub(self, x, y):
    return x - y
  
  def mul(self, x, y):
    return x * y
  


  def shape(self, x):
    return np.shape(x)

  def elt_type(self, x):
    return x.dtype if hasattr(x, 'dtype') else type(x)

  def alloc_array(self, elt_type, dims):
    return np.zeros(dims, dtype = elt_type)

  def shift_array(self, arr, offset):
    return arr[offset:]

  def ravel(self, arr):
    return np.ravel(arr)
  
  def index(self, arr, idx):
    return arr[idx]

  def tuple(self, elts):
    return tuple(elts)
  
  def tuple_elts(self, t):
    if isinstance(t, tuple):
      return t
    else:
      return (t,) 
  
  def concat_tuples(self, t1, t2):
    return tuple(t1) + tuple(t2)

  def setidx(self, arr, idx, v):
    arr[idx] = v

  def loop(self, start_idx, stop_idx, body):
    for i in xrange(start_idx, stop_idx):
      body(i)

  
  class Accumulator:
    def __init__(self, value):
      self.value = value

    def get(self):
      return self.value

    def update(self, new_value):
      self.value = new_value

  def accumulate_loop(self, start_idx, stop_idx, body, init):
    acc = self.Accumulator(init)
    for i in xrange(start_idx, stop_idx):
      body(acc, i)
    return acc.get()

  def check_equal_sizes(self, sizes):
    assert len(sizes) > 0
    first = sizes[0]
    assert all(sz == first for sz in sizes[1:])

  def slice_value(self, start, stop, step):
    return slice(start, stop, step)

  def invoke(self, fn, args):
    return eval_fn(fn, args)

  none = None
  null_slice = slice(None, None, None)
  def identity_function(self, x):
    return x

adverb_evaluator = InterpSemantics()

class ReturnValue(Exception):
  def __init__(self, value):
    self.value = value

class ClosureVal:
  def __init__(self, fn, fixed_args):
    self.fn = fn
    self.fixed_args = tuple(fixed_args)

  def __call__(self, args):
    if isinstance(args, ActualArgs):
      args = args.prepend_positional(self.fixed_args)
    else:
      args = self.fixed_args + tuple(args)
    return eval_fn(self.fn, args)

def eval_fn(fn, actuals):
  if isinstance(fn, np.dtype):
    return fn.type(*actuals)
  elif isinstance(fn, syntax.TypedFn):
    assert len(fn.arg_names) == len(actuals), \
      "Wrong number of args, expected %s but given %s" % \
      (fn.arg_names, actuals)
    env = {}

    for (k,v) in zip(fn.arg_names, actuals):
      env[k] = v
  elif isinstance(fn, syntax.Fn):
    # untyped functions have a more complicated args object
    # which deals with named args, variable arity, etc..
    env = fn.args.bind(actuals)
  elif isinstance(fn, ClosureVal):
    return fn(actuals)
  else:
    return fn(*actuals)

  def eval_args(args):
    if isinstance(args, (list, tuple)):
      return map(eval_expr, args)
    else:
      return args.transform(eval_expr)

  def eval_if_expr(maybe_expr):
    return eval_expr(maybe_expr) if isinstance(maybe_expr, syntax.Expr) else maybe_expr
  
  def eval_expr(expr):
    # print ">>", expr
    if hasattr(expr, 'wrapper'):
      expr = expr.wrapper
    assert isinstance(expr, syntax.Expr), "Not an expression-- %s : %s" % \
         (expr, type(expr))
    def expr_Const():
      return expr.value

    def expr_Attribute():
      value = eval_expr(expr.value)
      if expr.name == 'offset':
        if value.base is None:
          return 0
        else:
          return value.ctypes.data - value.base.ctypes.data
      else:
        return getattr(value, expr.name)

    def expr_ArrayView():
      data = eval_expr(expr.data)
      shape  = eval_expr(expr.shape)
      strides = eval_expr(expr.strides)
      offset = eval_expr(expr.offset)
      dtype = expr.type.elt_type.dtype
      return np.ndarray(shape = shape, 
                        offset = offset, 
                        buffer = data, 
                        strides = strides, 
                        dtype = np.dtype(dtype))
      
      
    def expr_Array():
      elt_values = map(eval_expr, expr.elts)
      return np.array(elt_values)

    def expr_Index():
      array = eval_expr(expr.value)
      index = eval_expr(expr.index)
      return array[index]

    def expr_PrimCall():
      return expr.prim.fn (*eval_args(expr.args))
    
    def expr_Slice():
      return slice(eval_expr(expr.start), eval_expr(expr.stop),
                   eval_expr(expr.step))

    def expr_Var():
      return env[expr.name]

    def expr_Call():
      fn = eval_expr(expr.fn)
      arg_values = eval_args(expr.args)
      return eval_fn(fn, arg_values)

    def expr_Closure():
      if isinstance(expr.fn, (syntax.Fn, syntax.TypedFn)):
        fundef = expr.fn
      else:
        assert isinstance(expr.fn, str)
        fundef = syntax.Fn.registry[expr.fn]
      closure_arg_vals = map(eval_expr, expr.args)
      return ClosureVal(fundef, closure_arg_vals)

    def expr_Fn():
      return ClosureVal(expr, [])

    def expr_TypedFn():
      return ClosureVal(expr, [])

    def expr_Cast():
      x = eval_expr(expr.value)
      t = expr.type
      assert isinstance(t, ScalarT)
      # use numpy's conversion function
      return t.dtype.type(x)

    def expr_Struct():
      assert expr.type, "Expected type on %s!" % expr
      assert isinstance(expr.type, StructT), \
          "Expected %s : %s to be a struct" % (expr, expr.type)
      elts = map(eval_expr, expr.args)
      return expr.type.ctypes_repr(elts)

    def expr_Tuple():
      return tuple(map(eval_expr, expr.elts))
    
    def expr_TupleProj():
      return eval_expr(expr.tuple)[expr.index]

    def expr_ClosureElt():
      assert isinstance(expr.closure, syntax.Expr), \
          "Invalid closure expression-- %s : %s" % \
          (expr.closure, type(expr.closure))
      clos = eval_expr(expr.closure)
      return clos.fixed_args[expr.index]

    def expr_Range():
      return np.arange(eval_expr(expr.start), eval_expr(expr.stop), eval_expr(expr.step))
    
    def expr_Len():
      return len(eval_expr(expr.value))
    
    def expr_IndexMap():
      fn = eval_expr(expr.fn)
      shape = eval_expr(expr.shape)
      ranges = [xrange(n) for n in shape]
      def wrap_idx(idx):
        if len(idx) == 1:
          idx = idx[0]
        return eval_fn(fn, (idx,))
      elts = [wrap_idx(idx) for idx in itertools.product(*ranges)]
      return np.array(elts).reshape((shape))
    
    def expr_IndexReduce():
      fn = eval_expr(expr.fn)
      combine = eval_expr(expr.combine)
      shape = eval_expr(expr.shape)
      ranges = [xrange(n) for n in shape]
        
      acc = eval_if_expr(expr.init)
      for idx in itertools.product(*ranges):
        if len(idx) == 1:
          idx = idx[0]
        elt = eval_fn(fn, (idx,))
        if acc is None:
          acc = elt 
        else:
          elt = eval_fn(combine, (acc, elt))
      return elt 
    
    def expr_Map():
      fn = eval_expr(expr.fn)
      args = eval_args(expr.args)
      axis = eval_if_expr(expr.axis)
      return adverb_evaluator.eval_map(fn, args, axis)

    def expr_AllPairs():
      fn = eval_expr(expr.fn)
      x,y = eval_args(expr.args)
      axis = eval_if_expr(expr.axis)
      return adverb_evaluator.eval_allpairs(fn, x, y, axis)

    def expr_Reduce():
      
      map_fn = eval_expr(expr.fn)
      combine_fn = eval_expr(expr.combine)
      args = eval_args(expr.args)
      init = eval_expr(expr.init) if expr.init else None
      axis = eval_if_expr(expr.axis)
      if axis is None:
        args = [np.ravel(x) for x in args]
        axis = 0
      return adverb_evaluator.eval_reduce(map_fn, combine_fn, init, args, axis)

    def expr_Scan():
      map_fn = eval_expr(expr.fn)
      combine = eval_expr(expr.combine)
      emit = eval_expr(expr.emit)
      args = eval_args(expr.args)
      init = eval_expr(expr.init)
      axis = eval_if_expr(expr.axis)
      return adverb_evaluator.eval_scan(map_fn, combine, emit, init, args, axis)
  
        
    result = dispatch(expr, 'expr')
    # we don't support python function's inside parakeet,
    # they have to be translated into Parakeet functions
    if isinstance(result, types.FunctionType):
      fundef = ast_conversion.translate_function_value(result)
      return ClosureVal(fundef, fundef.python_nonlocals())
    else:
      return result

  def eval_merge_left(phi_nodes):
    for result, (left, _) in phi_nodes.iteritems():
      env[result] = eval_expr(left)

  def eval_merge_right(phi_nodes):
    for result, (_, right) in phi_nodes.iteritems():
      env[result] = eval_expr(right)

  def assign(lhs, rhs, env):
    if isinstance(lhs, syntax.Var):
      env[lhs.name] = rhs
    elif isinstance(lhs, syntax.Tuple):
      assert isinstance(rhs, tuple)
      for (elt, v) in zip(lhs.elts, rhs):
        assign(elt, v, env)
    elif isinstance(lhs, syntax.Index):
      arr = eval_expr(lhs.value)
      idx = eval_expr(lhs.index)
      arr[idx] = rhs

  def eval_stmt(stmt):
    if isinstance(stmt, syntax.Return):
      v = eval_expr(stmt.value)
      raise ReturnValue(v)
    elif isinstance(stmt, syntax.Assign):
      value = eval_expr(stmt.rhs)
      assign(stmt.lhs, value, env)

    elif isinstance(stmt, syntax.If):
      cond_val = eval_expr(stmt.cond)
      if cond_val:
        eval_block(stmt.true)
        eval_merge_left(stmt.merge)
      else:
        eval_block(stmt.false)
        eval_merge_right(stmt.merge)

    elif isinstance(stmt, syntax.While):
      eval_merge_left(stmt.merge)
      while eval_expr(stmt.cond):
        eval_block(stmt.body)
        eval_merge_right(stmt.merge)
    elif isinstance(stmt, syntax.ForLoop):
      start = eval_expr(stmt.start)
      stop = eval_expr(stmt.stop)
      step = eval_expr(stmt.step)
      eval_merge_left(stmt.merge)
      for i in xrange(start, stop, step):
        env[stmt.var.name] = i
        eval_block(stmt.body)
        eval_merge_right(stmt.merge)
    else:
      raise RuntimeError("Not implemented: %s" % stmt)

  def eval_block(stmts):
    for stmt in stmts:
      eval_stmt(stmt)

  try:
   
    eval_block(fn.body)

  except ReturnValue as r:
    return r.value
  except:
    raise

def run_python_fn(python_fn, args, kwds):
  untyped  = ast_conversion.translate_function_value(python_fn)
  # should eventually roll this up into something cleaner, since
  # top-level functions are really acting like closures over their
  # global dependencies
  global_args = [python_fn.func_globals[n] for n in untyped.nonlocals]
  all_positional = global_args + list(args)
  actuals = args.FormalArgs(all_positional, kwds)
  return eval_fn(untyped, actuals)



from testing_helpers import expect, run_local_tests
import parakeet 
import numpy as np 

def varargs_return(*x):
  return x

def test_varargs_return():
  expect(varargs_return, [1,2], (1,2))

def varargs_add(*x):
  return x[0] + x[1]

def test_varargs_add():
  expect(varargs_add, [1,2], 3)

def call_varargs_add(x,y):
  local_tuple = (x,y)
  return varargs_add(*local_tuple)

def test_call_varargs_add():
  expect(call_varargs_add, [1,2], 3)
  expect(call_varargs_add, [True,2.0], 3.0)

def add_defaults(x = 1, y = 2):
    return x + y

def test_add_defaults():
  expect(add_defaults, [], 3)
  expect(add_defaults, [10], 12)
  expect(add_defaults, [10, 20], 30)
  expect(add_defaults, [10, 20.0], 30.0)

def call_add_defaults():
  return add_defaults(10)

def test_call_add_defaults():
  expect(call_add_defaults, [], 12)

def call_add_defaults_with_names():
  return add_defaults(y = 10, x = 20)

def test_call_defaults_with_names():
  expect(call_add_defaults_with_names, [], 30)

def sub(x,y):
  return x - y

def call_pos_with_names():
  return sub(y = 10, x = 20)

def test_call_pos_with_names():
  expect(call_pos_with_names, [], 10)
  
def tuple_default(x = (1,2)):
  return x[0] + x[1]

def test_tuple_default():
  expect(tuple_default, [], 3)
  expect(tuple_default, [(1,3)], 4)

def default_closure(x = 1, y=(1,2)):
  def inner(z, r = 0):
    return y[0] + y[1] + z + r
  return inner(x)

def test_default_closure(): 
  expect(default_closure, [], 4)
  expect(default_closure, [10], 13)
  expect(default_closure, [10, (20,30)], 60)

def default_none(x, y = None):
  if y is None:
    return x * 2
  else:
    return x + y

def test_default_none():
  expect(default_none, [10], 20)
  expect(default_none, [10, None], 20)
  expect(default_none, [10, 1], 11)

def add1(x):
  return x + 1

def add2(x): 
  return x + 2 


def fn_as_default(f = add1):
  return f(1)

def test_fn_as_default():
  expect(fn_as_default, [], 2)
  expect(fn_as_default, [add2], 3)

def lambda_closure(x):
  y = 1.0
  def g(f, x):
    return f(x)
  return g(lambda z: z + y, x)

def test_lambda_closure():
  expect(lambda_closure, [1], 2.0) 

if __name__ == '__main__':
  run_local_tests()



import numpy as np
import parakeet
import testing_helpers

#
#def test_col_sum():
#  m = 500
#  n = 5000
#  X = np.random.randn(m,n)
#  print "Reduce(Map)"
#  testing_helpers.timed_test(parakeet.sum, [X], np.sum, [X,0],
#                             min_speedup = 0.1)

def each_col_sum(X):
  return parakeet.each(parakeet.sum, X)

def test_each_col_sum():
  m = 1000
  n = 1000
  X = np.random.random((m,n))
  Xt = X.T
  # summing rows of Xt is same as summing cols of original
  testing_helpers.timed_test(each_col_sum, [Xt], np.sum, [Xt,1],
                             min_speedup = 0.05)

if __name__ == '__main__':
  testing_helpers.run_local_tests()




_type_mapping = {}
_typeof_functions = {}

def register(python_types, parakeet_type, typeof = None):
  """
  Map each python type to either a parakeet type or a function that returns a
  parakeet type
  """

  if typeof is None:
    typeof = lambda _: parakeet_type

  if not isinstance(python_types, (list, tuple)):
    python_types = [python_types]

  for python_type in python_types:
    _typeof_functions[python_type] = typeof
    _type_mapping[python_type] = parakeet_type

def equiv_type(python_type):
  assert python_type in _type_mapping, \
      "No type mapping found for %s" % python_type
  return _type_mapping[python_type]

def typeof(python_value):
  import decorators 
  if isinstance(python_value, decorators.jit):
    python_value = python_value.f 
  python_type = type(python_value)
  assert python_type in _typeof_functions, \
      "Don't know how to convert value %s : %s" % (python_value, python_type)
  return _typeof_functions[python_type](python_value)

def from_python(python_value):
  """
  Look up the ctypes representation of the corresponding parakeet type and call
  the converter with the ctypes class and python value
  """

  parakeet_type = typeof(python_value)
  result = parakeet_type.from_python(python_value)
  return result

def to_python(internal_value, parakeet_type):
  result = parakeet_type.from_python(internal_value)
  return result




# Copied from Russel's TreeLike but siginificantly 
# stripped down to fit in my underpowered brain 
 
import copy 
from itertools import izip 


_members_cache = {}  
_mro_cache = {}
_reversed_mro_cache = {}
  

class Node(object):
  
  @classmethod
  def get_mro(klass):
    class_name = klass.__name__ 
    if class_name in _mro_cache:
      return _mro_cache[class_name]
    else:
      mro = klass.mro()
      rev_mro = list(reversed(mro))
      _mro_cache[class_name] = mro
      _reversed_mro_cache[class_name] = rev_mro
      return mro 
    
  @classmethod
  def get_reverse_mro(klass):
    class_name = klass.__name__
    if class_name in _reversed_mro_cache:
      return _reversed_mro_cache[class_name]
    else:
      mro = klass.mro()
      rev_mro = list(reversed(mro))
      _mro_cache[class_name] = mro
      _reversed_mro_cache[class_name] = rev_mro
      return rev_mro 
      
      
  @classmethod
  def members(klass):
    """
    Walk through classes in mro order, accumulating member names.
    """
    class_name = klass.__name__
    if class_name  in _members_cache:
      return _members_cache[class_name]
    
    m = []
    for c in klass.get_mro():
      curr_members = getattr(c, '_members', []) 
      for name in curr_members:
        if name not in m:
          m.append(name)  
    _members_cache[class_name] = m
    return m
  
  def iteritems(self):
    for k in self.members():
      yield (k, getattr(self, k, None))
      
  def itervalues(self):
    for (_,v) in self.iteritems():
      yield v 
  
  def items(self):
    [(k,getattr(self,k)) for k in self.members()]
  
  def __init__(self, *args, **kw):
    members = self.members()
    n_args = len(args)
    n_members = len(members)
    class_name = self.__class__.__name__ 
    self_dict = self.__dict__
    if n_args == n_members:
      assert len(kw) == 0
      for (k,v) in zip(members,args):
        self_dict[k] = v
    elif n_args < n_members:
      for field in members:
        self_dict[field] = kw.get(field)
      
      for field, value in izip(members, args):
        self_dict[field] = value
        
      for (k,v) in kw.iteritems():
        assert k in members, \
          "Keyword argument '%s' not recognized for %s: %s" % \
          (k, self.node_type(), members)
    else:
      raise Exception('Too many arguments for %s, expected %s' % \
                      (class_name, members))
       
    # it's more common to not define a node initializer, 
    # so add an extra check to avoid having to always
    # traverse the full class hierarchy 
    if hasattr(self, 'node_init'):
      for C in _reversed_mro_cache[class_name]:
        if 'node_init' in C.__dict__:
          C.node_init(self)

  def __hash__(self):
    # print "Warning: __hash__ not implemented for %s" % self
    hash_values = []
    for m in self.members():
      v = getattr(self, m)
      if isinstance(v, (list, tuple)):
        v = tuple(v)
      hash_values.append(v)
    return hash(tuple(hash_values))
  
  def eq_members(self, other):
    for (k,v) in self.iteritems():
      if not hasattr(other, k):
        return False
      if getattr(other, k) != v:
        return False
    return True 
  
  def __eq__(self, other):
    return other.__class__ is  self.__class__ and self.eq_members(other)

  def __ne__(self, other):
    return not self == other 
  
  @classmethod
  def node_type(cls):
    return cls.__name__
  
  def clone(self, **kwds):
    cloned = copy.deepcopy(self)
    for (k,v) in kwds.values():
      setattr(cloned, k, v)
    return cloned 
    
  def __str__(self):
    member_strings = []
    for (k,v) in self.iteritems():
      member_strings.append("%s = %s" % (k, v)) 
    return "%s(%s)" % (self.node_type(), ", ".join(member_strings))
  
  def __repr__(self):
    return self.__str__()


import numpy as np

import parakeet
import testing_helpers

from parakeet import config, each, syntax
from parakeet.pipeline import lowering
from parakeet.syntax_visitor import SyntaxVisitor
from testing_helpers import expect

config.opt_fusion = True 
config.opt_copy_elimination = True 
config.opt_inline = True 
config.opt_licm = True 


def A(x):
  return x + 1

def B(x):
  return A(x)

def C(x):
  return B(x)

def simple_args(exprs):
  return all(simple_expr(e) for e in exprs)

def simple_primcall(e):
  return isinstance(e, syntax.PrimCall) and simple_args(e.args)

def simple_expr(expr):
  return isinstance(expr, (syntax.Const, syntax.Var)) or \
         simple_primcall(expr)

def simple_stmt(stmt):
  """
  Is this a statement from a simple straightline function?

  (can only contain scalar computations and no control flow)
  """
  if isinstance(stmt, syntax.Return):
    return simple_expr(stmt.value)
  elif isinstance(stmt, syntax.Assign):
    return simple_expr(stmt.rhs)
  else:
    return False

def simple_block(stmts):
  return all(simple_stmt(s) for s in stmts)

def simple_fn(fn):
  return simple_block(fn.body)

def test_inline():
  typed_fn = parakeet.typed_repr(C, [1])
  assert len(typed_fn.body) in [1,2], \
      "Expected body to be 1 or 2 statements, got %s" % typed_fn
  assert simple_fn(typed_fn)
  testing_helpers.expect(C, [1], 2)

def lots_of_arith(x):
  y = 4 * 1
  z = y + 1
  a = z / 5
  b = x * a
  return b

def test_simple_constant_folding():
  testing_helpers.expect(lots_of_arith, [1], 1)
  typed_fn = parakeet.typed_repr(lots_of_arith, [1])
  assert len(typed_fn.body) == 1, \
      "Insufficiently simplified body: %s" % typed_fn

def const_across_control_flow(b):
  if b:
    x = 1
  else:
    x = 1
  return x

def test_constants_across_control_flow():
  testing_helpers.expect(const_across_control_flow, [True], 1)
  typed_fn = parakeet.typed_repr(const_across_control_flow, [True])
  assert len(typed_fn.body) == 1, "Fn body too long: " + str(typed_fn.body)
  stmt = typed_fn.body[0]
  assert isinstance(stmt, syntax.Return)
  assert isinstance(stmt.value, syntax.Const)

def always_true_branch():
  x = 1 + 1
  if x == 2:
    res = 0 + 0
    return res
  else:
    res = 1 * 1 + 0
    return res

def test_always_true():
  testing_helpers.expect(always_true_branch, [], 0)
  typed_fn = parakeet.typed_repr(always_true_branch, [])
  assert len(typed_fn.body) == 1, "Fn body too long: " + str(typed_fn.body)
  stmt = typed_fn.body[0]
  assert isinstance(stmt, syntax.Return)
  assert isinstance(stmt.value, syntax.Const)

def always_false_branch():
  x = 1 + 2
  if x == 2:
    res = 1 * 0 + 0
    return res
  else:
    res = 0 + 1 * 1
    return res

def test_always_false():
  testing_helpers.expect(always_false_branch, [], 1)
  typed_fn = parakeet.typed_repr(always_false_branch, [])
  assert len(typed_fn.body) == 1, "Fn body too long: " + str(typed_fn.body)
  stmt = typed_fn.body[0]
  assert isinstance(stmt, syntax.Return)
  assert isinstance(stmt.value, syntax.Const)

def volatile_licm_mistake():
  i = 0
  x = [0]
  while i < 10:
    alloc = [1]
    if i == 5:
      x = alloc
    else:
      alloc[0] = 100
    i = i + 1
  return x

def test_volatile_licm_mistake():
  expect(volatile_licm_mistake, [], np.array([1]))

def g(x):
  def h(xi):
    return xi + 1.0
  return each(h,x)

def nested_add1(X):
  return each(g, X)

class CountLoops(SyntaxVisitor):
  def __init__(self):
    SyntaxVisitor.__init__(self)
    self.count = 0

  def visit_While(self, stmt):
    self.count += 1
    SyntaxVisitor.visit_While(self, stmt)

def count_loops(fn):
  Counter = CountLoops()
  Counter.visit_fn(fn)
  return Counter.count

def test_copy_elimination():
  x = np.array([[1,2,3],[4,5,6]])
  expect(nested_add1, [x], x + 1.0)
  typed_fn = parakeet.typed_repr(nested_add1, [x])
  lowered = lowering.apply(typed_fn)
  n_loops = count_loops(lowered)
  n_expected = 3 if config.opt_loop_unrolling else 2
  assert n_loops <= n_expected, \
      "Too many loops generated! Expected at most 2, got %d" % n_loops

def licm_nested_loops():
  
  total = 0
  a = [1,2,3,4,5]
  for _ in range(3):
    for i in range(2):
      for j in range(len(a)):
        a[j] *= 10
    total += a[1]
  return total   

def test_licm_nested_loops():
  testing_helpers.expect(licm_nested_loops, [], licm_nested_loops())
  
if __name__ == '__main__':
  testing_helpers.run_local_tests()



import args
import array_type
import syntax
import tuple_type

from core_types import Int64, Int32, Float32, Float64, Bool
from core_types import FloatT, IntT, BoolT, NoneType, ScalarT
from syntax import Const

def const_int(n, t = Int64):
  return Const(n, type = t)

def const_float(f, t = Float64):
  return Const(f, type = t)

def const_bool(b, t = Bool):
  return Const(b, type = t)

def zero(t):
  if isinstance(t, FloatT):
    x = 0.0
  elif isinstance(t, BoolT):
    x = False
  else:
    assert isinstance(t, IntT)
    x = 0
  return Const(x, type = t)

false = zero(Bool)
zero_i32 = zero(Int32)
zero_i64 = zero(Int64)
zero_f32 = zero(Float32)
zero_f64 = zero(Float64)

def one(t):
  if isinstance(t, FloatT):
    x = 1.0
  elif t.__class__ is BoolT:
    x = True
  else:
    assert isinstance(t, IntT)
    x = 1
  return Const(x, type = t)

true = one(Bool)
one_i32 = one(Int32)
one_i64 = one(Int64)
one_f32 = one(Float32)
one_f64 = one(Float64)

none_t = NoneType
none = Const(None, type = none_t)

slice_none_t = array_type.make_slice_type(none_t, none_t, none_t)
slice_none = syntax.Slice(none, none, none, type = slice_none_t)

def is_python_int(x):
  return isinstance(x, (int, long))

def is_python_float(x):
  return isinstance(x, float)

def is_python_bool(x):
  return isinstance(x, bool)

def is_python_scalar(x):
  return isinstance(x,  (bool, int, long, float))

def is_python_constant(x):
  if isinstance(x, tuple):
    return all(map(is_python_constant, x))
  else:
    return x is None or is_python_scalar(x)

def const_scalar(x):
  if is_python_int(x):
    return const_int(x)
  elif isinstance(x, bool):
    return const_bool(x)
  else:
    assert isinstance(x, float)
    return const_float(x)

def make_tuple(elts):
  elt_types = get_types(elts)
  tuple_t = tuple_type.make_tuple_type(elt_types)
  return syntax.Tuple(elts, type = tuple_t)

def const_tuple(*python_values):
  return make_tuple(map(const, python_values))

def const(x):
  if is_python_scalar(x):
    return const_scalar(x)
  elif isinstance(x, tuple):
    return const_tuple(*map(const, x))
  elif isinstance(x, syntax.Expr):
    return x
  else:
    assert x is None, \
        "Can't convert Python value %s into a Parakeet constant" % x
    return none

def unwrap_constant(x):
  if isinstance(x, syntax.Expr):
    assert x.__class__ is Const, \
        "Expected constant, got %s" % (x,)
    return x.value
  else:
    assert is_python_constant(x)
    return x

def wrap_if_constant(x):
  """
  If given value isn't already an expression turn it into one with the const
  helper
  """

  if is_python_constant(x):
    return const(x)
  else:
    assert isinstance(x, syntax.Expr), "Expected expression, got " + str(x)
    return x

def wrap_constants(xs):
  return map(wrap_if_constant, xs)

def wrap_var(x):
  if isinstance(x, str):
    return syntax.Var(x)
  else:
    assert isinstance(x, syntax.Var), "Expected a variable, got %s" % x

def wrap_vars(xs):
  return map(wrap_var, xs)

def get_type(expr):
  return expr.type

def get_types(exprs):
  if isinstance(exprs, args.ActualArgs):
    return exprs.transform(get_type)
  else:
    return [expr.type for expr in exprs]

def is_zero(expr):
  return expr.__class__ is Const and expr.value == 0

def is_one(expr):
  return expr.__class__ is Const and expr.value == 1

def is_false(expr):
  return expr.__class__ is Const and expr.value == False

def is_true(expr):
  return expr.__class__ is Const and expr.value == True

def is_none(expr):
  return expr.__class__ is Const and expr.value == None

def is_constant(expr):
  return expr.__class__ is Const

def all_constants(exprs):
  return all(map(is_constant, exprs))

def collect_constant(expr):
  return expr.value

def collect_constants(exprs):
  return map(collect_constant, exprs)

def is_scalar(expr):
  return isinstance(expr.type, ScalarT)

def all_scalars(exprs):
  return all(map(is_scalar, exprs))

def is_identity_fn(fn):
  return len(fn.arg_names) == 1 and len(fn.body) == 1 and \
         fn.body[0].__class__ is syntax.Return and \
         fn.body[0].value.__class__ is syntax.Var and \
         fn.body[0].value.name == fn.arg_names[0]
         
def gen_arg_names(n, base_names):
  results = []

  m = len(base_names)
  for i in xrange(n):
    curr = base_names[i % m]
    cycle = i / m
    if cycle > 0:
      curr = "%s_%d" % (curr, cycle+1)
    results.append(curr)
  return results

def gen_data_arg_names(n):
  return gen_arg_names(n, ['x', 'y', 'z', 'a', 'b', 'c', 'd'])

def gen_fn_arg_names(n):
  return gen_arg_names(n, ['f', 'g', 'h', 'p', 'q', 'r', 's'])




import numpy as np
import testing_helpers

from parakeet import each

def add1(xi):
  return xi + 1

def add11d(x):
  return each(add1, x)

int_vec = np.arange(128, dtype=np.float).reshape(16,8)

def test_add1():
  result = each(add11d, int_vec)
  expected = int_vec + 1
  assert testing_helpers.eq(result, expected), \
      "Expected %s, got %s" % (expected, result)

if __name__ == '__main__':
  testing_helpers.run_local_tests()



import shape 
import syntax  


from shape_inference import shape_env
from transform import Transform

class ShapeElimination(Transform): 
  def pre_apply(self, fn):
    self.shape_env = shape_env(fn)

#    print "Inferred shapes"
#    for (k,v) in sorted(self.shape_env.items()):
#      print "  %s => %s" % (k,v)
  def transform_Var(self, expr):
    if expr.name in self.shape_env:
      v = self.shape_env[expr.name]
      if v.__class__ is shape.Const:
        return  syntax.Const(value = v.value, type = expr.type)
    return expr
  
  def transform_lhs(self, lhs):
    return lhs


import numpy as np
import parakeet 

from testing_helpers import expect, run_local_tests 


int_vec = np.array([1,2,3])
float_vec = np.array([1.0,2.0,3.0])
bool_vec = np.array([True,False])

int_mat = np.array([int_vec, int_vec])
float_mat = np.array([float_vec, float_vec])
bool_mat = np.array([bool_vec, bool_vec])

def test_ones():
  expect(parakeet.ones, [(10,)], np.ones(10))
  expect(parakeet.ones, [(10,2)], np.ones((10,2)))  
  expect(parakeet.ones, [(10,), np.int64], np.ones(10, dtype=np.int64))
  expect(parakeet.ones, [(10,2), np.int64], np.ones((10,2), dtype=np.int64))
  expect(parakeet.ones, [(10,), np.float64], np.ones(10, dtype=np.float64))
  expect(parakeet.ones, [(10,2), np.float64], np.ones((10,2), dtype=np.float64))
  expect(parakeet.ones, [(10,2,2,2), np.float64], np.ones((10,2,2,2), dtype=np.float64))

def test_ones_like():
  expect(parakeet.ones_like, [int_vec], np.ones_like(int_vec))
  expect(parakeet.ones_like, [int_mat], np.ones_like(int_mat))
  expect(parakeet.ones_like, [float_vec], np.ones_like(float_vec))
  expect(parakeet.ones_like, [float_mat], np.ones_like(float_mat))
  expect(parakeet.ones_like, [bool_vec], np.ones_like(bool_vec))
  expect(parakeet.ones_like, [bool_mat], np.ones_like(bool_mat))
  expect(parakeet.ones_like, [bool_vec, np.dtype('float')], np.ones_like(bool_vec, dtype = np.dtype('float')))
  expect(parakeet.ones_like, [bool_mat, np.dtype('float')], np.ones_like(bool_mat, dtype = np.dtype('float')))


def test_zeros():
  expect(parakeet.zeros, [(10,)], np.zeros(10))
  expect(parakeet.zeros, [(10,2)], np.zeros((10,2)))  
  expect(parakeet.zeros, [(10,), np.int64], np.zeros(10, dtype=np.int64))
  expect(parakeet.zeros, [(10,2), np.int64], np.zeros((10,2), dtype=np.int64))
  expect(parakeet.zeros, [(10,), np.float64], np.zeros(10, dtype=np.float64))
  expect(parakeet.zeros, [(10,2), np.float64], np.zeros((10,2), dtype=np.float64))
  expect(parakeet.zeros, [(10,2,2,2), np.float64], np.zeros((10,2,2,2), dtype=np.float64))

def test_zeros_like():
  expect(parakeet.zeros_like, [int_vec], np.zeros_like(int_vec))
  expect(parakeet.zeros_like, [int_mat], np.zeros_like(int_mat))
  expect(parakeet.zeros_like, [float_vec], np.zeros_like(float_vec))
  expect(parakeet.zeros_like, [float_mat], np.zeros_like(float_mat))
  expect(parakeet.zeros_like, [bool_vec], np.zeros_like(bool_vec))
  expect(parakeet.zeros_like, [bool_mat], np.zeros_like(bool_mat))
  expect(parakeet.zeros_like, [bool_vec, np.dtype('float')], np.zeros_like(bool_vec, dtype = np.dtype('float')))
  expect(parakeet.zeros_like, [bool_mat, np.dtype('float')], np.zeros_like(bool_mat, dtype = np.dtype('float')))


def test_empty():
  s = (20, 20, 3)
  x = parakeet.empty(s)
  assert x.shape == s
  
def test_empty_int():
  s = (2,2,2,2)
  x = parakeet.empty(s, dtype=np.uint8)
  assert x.shape == s
  assert x.dtype == np.uint8


if __name__ == '__main__':
  run_local_tests()



import array_type
import names

from syntax import AllPairs, Map
from syntax import Assign, Return, TypedFn, Var
from transform import Transform

class MapifyAllPairs(Transform):
  def transform_stmt(self, stmt):
    """
    Assume that all adverbs occur only at the outer level of bindings, so skip
    recursive evaluation of expressions
    """

    if stmt.__class__ is Assign and stmt.rhs.__class__ is AllPairs:
      stmt.rhs = self.transform_AllPairs(stmt.rhs)
    elif stmt.__class__ is Return and stmt.value.__class__ is AllPairs:
      stmt.value = self.transform_AllPairs(stmt.value)
    return stmt

  def transform_AllPairs(self, expr):
    """
    Transform each AllPairs(f, X, Y) operation into a pair of nested maps:
      def g(x_elt):
        def h(y_elt):
          return f(x_elt, y_elt)
      return map(g, X)
    """

    if expr.out is not None:
      return expr

    # if the adverb function is a closure, give me all the values it
    # closes over
    closure_elts = self.closure_elts(expr.fn)
    n_closure_elts = len(closure_elts)

    # strip off the closure wrappings and give me the underlying TypedFn
    fn = self.get_fn(expr.fn)

    # the two array arguments to this AllPairs adverb
    x, y_outer = expr.args

    x_elt_name = names.fresh('x_elt')
    x_elt_t = fn.input_types[n_closure_elts]
    x_elt_var = Var(x_elt_name, type = x_elt_t)
    y_inner_name = names.fresh('y')
    y_inner = Var(y_inner_name, type = y_outer.type)

    inner_closure_args = []
    for (i, elt) in enumerate(closure_elts):
      t = elt.type
      if elt.__class__ is Var:
        name = names.refresh(elt.name)
      else:
        name = names.fresh('closure_arg%d' % i)
      inner_closure_args.append(Var(name, type = t))

    inner_arg_names = []
    inner_input_types = []
    type_env = {}

    for var in inner_closure_args + [y_inner, x_elt_var]:
      type_env[var.name] = var.type
      inner_arg_names.append(var.name)
      inner_input_types.append(var.type)

    inner_closure_rhs = self.closure(fn, inner_closure_args + [x_elt_var])

    inner_result_t = array_type.lower_rank(expr.type, 1)
    inner_fn = TypedFn(
      name = names.fresh('allpairs_into_maps_wrapper'),
      arg_names = tuple(inner_arg_names),
      input_types = tuple(inner_input_types),
      return_type = inner_result_t,
      type_env = type_env,
      body = [
        Return(Map(inner_closure_rhs,
                   args=[y_inner],
                   axis = expr.axis,
                   type = inner_result_t))
      ]
    )
    closure = self.closure(inner_fn, closure_elts + [y_outer])
    return Map(closure, [x], axis = expr.axis, type = expr.type)



import matplotlib.pylab as pylab
import numpy as np
import parakeet

from testing_helpers import eq, run_local_tests

size = 21
init = np.array(([0] * (size/2)) + [1] + ([0] * (size - size/2 - 1)))

def rule30(extended, i):
  a, b, c = extended[[i-1,i,i+1]]
  if ((a == 1 and b == 0 and c == 0) or
      (a == 0 and b == 1 and c == 1) or
      (a == 0 and b == 1 and c == 0) or
      (a == 0 and b == 0 and c == 1)):
    return 1
  else:
    return 0

plot = False

def test_rule30():
  output = init.copy()
  cur = init
  zero_array = np.array([0])
  indices = np.arange(1,size+1)
  for _ in range(size/2):
    extended = np.concatenate((zero_array, cur, zero_array))

    def run_rule30(i):
      return rule30(extended, i)
    parakeet_iter = parakeet.each(run_rule30, indices)
    cur = np.array(map(run_rule30, indices))
    assert eq(parakeet_iter, cur), \
       "Parakeet result (%s) didn't match Python result(%s)" % \
       (parakeet_iter, cur)
    output = np.vstack((output,cur))

  if plot:
    pylab.matshow(output,cmap = pylab.cm.gray)
    pylab.show()

if __name__ == '__main__':
  run_local_tests()



import names

import syntax
import syntax_helpers
from args import FormalArgs

class macro(object):
  def __init__(self, f, static_names = set([]), call_from_python = None):
    self.f = f
    self.static_names = static_names
    self.wrappers = {}
    self.call_from_python = call_from_python
    if hasattr(self.f, "__name__"):
      self.name = f.__name__
    else:
      self.name = "f"

  _macro_wrapper_cache = {}
  def _create_wrapper(self, n_pos, static_pairs, dynamic_keywords):
    args = FormalArgs()
    pos_vars = []
    keyword_vars = {}
    for i in xrange(n_pos):
      local_name = names.fresh("input_%d" % i)
      args.add_positional(local_name)
      pos_vars.append(syntax.Var(local_name))
  
    
    for visible_name in dynamic_keywords:
      local_name = names.fresh(visible_name)
      args.add_positional(local_name, visible_name)
      keyword_vars[visible_name] = syntax.Var(local_name)

    for (static_name, value) in static_pairs:
      if isinstance(value, syntax.Expr):
        assert isinstance(value, syntax.Const)
        keyword_vars[static_name] = value
      elif value is not None:
        assert syntax_helpers.is_python_constant(value), \
            "Unexpected type for static/staged value: %s : %s" % \
            (value, type(value))
        keyword_vars[static_name] = syntax_helpers.const(value)

    result_expr = self.f(*pos_vars, **keyword_vars)
    body = [syntax.Return(result_expr)]
    wrapper_name = "%s_wrapper_%d_%d" % (self.name, n_pos,
                                         len(dynamic_keywords))
    wrapper_name = names.fresh(wrapper_name)
    return syntax.Fn(name = wrapper_name, args = args, body = body)

  def as_fn(self):
    n_args = self.f.func_code.co_argcount
    n_default = 0 if not self.f.func_defaults else len(self.f.func_defaults)
    assert n_default == 0
    return self._create_wrapper(n_args,[],{})
    
  def __call__(self, *args, **kwargs):
    import run_function
    if self.call_from_python is None:
      n_pos = len(args)
      keywords = kwargs.keys()

      static_pairs = ((k,kwargs.get(k)) for k in self.static_names)
      dynamic_keywords = tuple(k for k in keywords
                               if k not in self.static_names)

      static_pairs = tuple(static_pairs)
      key = (n_pos, static_pairs, dynamic_keywords)

      if key in self.wrappers:
        untyped = self.wrappers[key]
      else:
        untyped = self._create_wrapper(n_pos, static_pairs, dynamic_keywords)
        self.wrappers[key] = untyped
      dynamic_kwargs = dict( (k, kwargs[k]) for k in dynamic_keywords)
      return run_function.run(untyped, *args, **dynamic_kwargs)
    else:
      return self.call_from_python(*args, **kwargs)

  def transform(self, args, kwargs = {}):
    for arg in args:
      assert isinstance(arg, syntax.Expr), \
          "Macros can only take syntax nodes as arguments, got %s" % (arg,)
    for (name,arg) in kwargs.iteritems():
      assert isinstance(arg, syntax.Expr), \
          "Macros can only take syntax nodes as arguments, got %s = %s" % \
          (name, arg)
    result = self.f(*args, **kwargs)
    assert isinstance(result, syntax.Expr), \
        "Expected macro %s to return syntax expression, got %s" % \
        (self.f, result)
    return result

  def __str__(self):
    return "macro(%s)" % self.name

class staged_macro(object):
  def __init__(self, *static_names, **kwargs):
    self.static_names = tuple(static_names)

    self.call_from_python = kwargs.get('call_from_python')
    assert kwargs.keys() in [[], ['call_from_python']], \
        "Unknown keywords: %s" % kwargs.keys()

  def __call__(self, fn):
    return macro(fn, self.static_names,
                 call_from_python = self.call_from_python)
   
class jit:
  def __init__(self, f):
    self.f = f

  def __call__(self, *args, **kwargs):
    import run_function
    return run_function.run(self.f, *args, **kwargs)




import numpy as np
import types 

import core_types 
import prims  
import type_conv

from array_type import make_array_type, ArrayT
from closure_type import make_closure_type, ClosureT
from core_types import NoneT, NoneType, TypeValueT, from_dtype
from decorators import jit, macro 
from tuple_type import make_tuple_type, TupleT 

type_conv.register(type(None), NoneT, lambda _: NoneType)
type_conv.register([np.dtype], TypeValueT, lambda dt: TypeValueT(from_dtype(dt))) 

def typeof_type(t):
  assert hasattr(t, 'dtype'), "Can only convert numpy types"
  dt = t(0).dtype 
  pt = from_dtype(dt)
  return TypeValueT(pt) 
type_conv.register(types.TypeType, TypeValueT, typeof_type)

def typeof_tuple(python_tuple):
  return make_tuple_type(map(type_conv.typeof, python_tuple))

type_conv.register(types.TupleType, TupleT, typeof_tuple)

def typeof_array(x):
  x = np.asarray(x)
  elt_t = core_types.from_dtype(x.dtype)
  rank = len(x.shape)
  return make_array_type(elt_t, rank)

type_conv.register((np.ndarray, list, xrange), ArrayT, typeof_array)

def typeof_prim(p):
  untyped_fn = prims.prim_wrapper(p)
  return make_closure_type(untyped_fn, [])

type_conv.register(prims.class_list, ClosureT, typeof_prim)

def typeof_fn(f):
  import ast_conversion
  untyped_fn = ast_conversion.translate_function_value(f)
  closure_args = untyped_fn.python_nonlocals()
  closure_arg_types = map(type_conv.typeof, closure_args)
  return make_closure_type(untyped_fn, closure_arg_types)

type_conv.register(types.FunctionType, ClosureT, typeof_fn)
type_conv.register(jit, ClosureT, typeof_fn)
type_conv.register(macro, ClosureT, typeof_fn)
type_conv.register(types.BuiltinFunctionType, ClosureT, typeof_fn)




import numpy as np

# partial mapping, since ctypes doesn't support
# complex numbers
bool8 = np.dtype('bool8')
int8 = np.dtype('int8')
uint8 = np.dtype('uint8')
int16 = np.dtype('int16')
uint16 = np.dtype('uint16')
int32 = np.dtype('int32')
uint32 = np.dtype('uint32')
int64 = np.dtype('int64')
uint64 = np.dtype('uint64')
  
  
float32 = np.dtype('float32')
float64 = np.dtype('float64')
complex64 = np.dtype('complex64')
complex128 = np.dtype('complex128')
  
def is_float(dtype):
  return dtype.type in np.sctypes['float']

def is_signed(dtype):
  return dtype.type in np.sctypes['int']
  
def is_unsigned(dtype):
  return dtype.type in np.sctypes['uint']

def is_complex(dtype):
  return dtype.type in np.sctypes['complex']
   
def is_bool(dtype):
  return dtype == np.bool8
   
def is_int(dtype):
  return is_bool(dtype) or is_signed(dtype) or is_unsigned(dtype)
  

import ctypes 
_to_ctypes = {
  bool8 : ctypes.c_bool, 
  
  int8  : ctypes.c_int8, 
  uint8 : ctypes.c_uint8, 
  
  int16 : ctypes.c_int16, 
  uint16 : ctypes.c_uint16,
  
  int32 : ctypes.c_int32, 
  uint32 : ctypes.c_uint32,
  
  int64 : ctypes.c_int64, 
  uint64 : ctypes.c_uint64, 
  
  float32 : ctypes.c_float, 
  float64 :  ctypes.c_double, 
}

def to_ctypes(dtype):
  """
  Give the ctypes representation for each numpy scalar type. 
  Beware that complex numbers have no assumed representation 
  and thus aren't valid arguments to this function. 
  """
  if dtype in _to_ctypes:
    return _to_ctypes[dtype]
  else:
    raise RuntimeError("No conversion from %s to ctypes" % dtype)
  



import ctypes
import numpy as np

import dtypes
import type_conv

from node import Node

class TypeFailure(Exception):
  def __init__(self, msg):
    self.msg = msg

class IncompatibleTypes(Exception):
  def __init__(self, t1, t2):
    self.t1 = t1
    self.t2 = t2

  def __repr__(self):
    return "IncompatibleTypes(%s, %s)" % (self.t1, self.t2)

  def __str__(self):
    return repr(self)

class Type(Node):
  def combine(self, other):
    raise IncompatibleTypes(self, other)

  def __hash__(self):
    assert False, "Hash function not implemented for type %s" % (self,)

  def __eq__(self, _):
    assert False, "Equality not implemented for type %s" % (self,)

class AnyT(Type):
  """top of the type lattice, absorbs all types"""

  def combine(self, other):
    return self

  def __eq__(self, other):
    return isinstance(other, AnyT)

# since there's only one Any type, just create an instance of the same name
Any = AnyT()

class UnknownT(Type):
  """Bottom of the type lattice, absorbed by all other types"""

  _members = []
  def  combine(self, other):
    return other

  def __eq__(self, other):
    return isinstance(other, UnknownT)

#single instance of the Unknown type with same name
Unknown = UnknownT()


class FnT(Type):
  """Type of a typed function"""

  def __init__(self, input_types, return_type):
    self.input_types = tuple(input_types )
    self.return_type = return_type

  def __str__(self):
    input_str = ", ".join(str(t) for t in self.input_types)
    return "(%s)->%s" % (input_str, self.return_type)

  def __repr__(self):
    return str(self)

  def __eq__(self, other):
    return isinstance(other, FnT) and \
           self.return_type == other.return_type and \
           len(self.input_types) == len(other.input_types) and \
           all(t1 == t2 for (t1, t2) in
               zip(self.input_types, other.input_types))

  def combine(self, other):
    if self == other:
      return self
    else:
      raise IncompatibleTypes(self, other)

  def __hash__(self):
    return hash(self.input_types + (self.return_type,))

_fn_type_cache = {}
def make_fn_type(input_types, return_type):
  input_types = tuple(input_types)
  key = input_types, return_type
  if key in _fn_type_cache:
    return _fn_type_cache[key]
  else:
    t = FnT(input_types, return_type)
    _fn_type_cache[key] = t
    return t


def combine_type_list(types):
  common_type = Unknown

  for t in types:
    common_type = common_type.combine(t)
  return common_type

class ConcreteT(Type):
  """
  Type which actually have some corresponding runtime values, as opposed to
  "Any" and "Unknown"
  """

  def from_python(self, py_val):
    return self.ctypes_repr(py_val)

  def to_python(self, internal):
    return internal

# None is fully described by its type, so the
# runtime representation can just be the number zero

class NoneT(ConcreteT):
  _members = []
  rank = 0
  ctypes_repr = ctypes.c_int64

  def from_python(self, val):
    assert val is None
    return ctypes.c_int64(0)

  def to_python(self, obj):
    assert obj == ctypes.c_int64(0), \
        "Expected runtime representation of None to be 0, but got: %s" % obj
    return None

  def combine(self, other):
    if isinstance(other, NoneT):
      return self
    else:
      raise IncompatibleTypes(self, other)

  def __str__(self):
    return "NoneT"

  def __hash__(self):
    return 0

  def __eq__(self, other):
    return other.__class__ is NoneT
  
  def __repr__(self):
    return str(self)

NoneType = NoneT()



def is_struct(c_repr):
  return type(c_repr) == type(ctypes.Structure)

class FieldNotFound(Exception):
  def __init__(self, struct_t, field_name):
    self.struct_t = struct_t
    self.field_name = field_name
    
  def __str__(self):
    return "FieldNotFound(%s, %s)" % (self.struct_t, self.field_name)

class StructT(Type):
  """All concrete types excluding scalars and pointers"""

  # expect each child class to fill this list
  _fields_ = []

  _repr_cache = {}

  def field_type(self, name):
    for (field_name, field_type) in self._fields_:
      if field_name == name:
        return field_type
    raise FieldNotFound(self, name)

  def field_pos(self, name):
    for (i, (field_name, _)) in enumerate(self._fields_):
      if field_name == name:
        return i
    raise FieldNotFound(self, name)

  @property
  def ctypes_repr(self):
    if self in self._repr_cache:
      return self._repr_cache[self]

    ctypes_fields = []

    for (field_name, parakeet_field_type) in self._fields_:
      field_repr = parakeet_field_type.ctypes_repr
      # nested structures will be heap allocated
      if isinstance(parakeet_field_type,  StructT):
        ptr_t = ctypes.POINTER(field_repr)
        ctypes_fields.append( (field_name, ptr_t) )
      else:
        ctypes_fields.append( (field_name, field_repr) )

    class Repr(ctypes.Structure):
      _fields_ = ctypes_fields
    Repr.__name__ = self.node_type() +"_Repr"
    self._repr_cache[self] = Repr
    return Repr

###################################################
#                                                 #
#             SCALAR NUMERIC TYPES                #
#                                                 #
###################################################

###################################################
# helper functions to implement properties of Python scalar objects
###################################################
def always_zero(x):
  return type(x)(0)

def identity(x):
  return x

class TypeValueT(ConcreteT):
  """
  Materialization of a type into a value 
  """
  rank = 0
  _members = ['type']
  
  def __str__(self):
    return "TypeValue(%s)" % self.type 
  
  def __hash__(self):
    return hash(self.type)
  
  def __eq__(self, other):
    return other.__class__ is TypeValueT and self.type == other.type
  
  _type_to_id = {}
  _id_to_type = {} 
  def from_python(self, py_type):
    if py_type in self._type_to_id:
      n = self._type_to_id[py_type]
    else:
      n = len(self._type_to_id)
      self._id_to_type[n] = py_type 
      self._type_to_id[py_type] = n
    return self.ctypes_repr(n)
  
  ctypes_repr = ctypes.c_int
  

# base class for all concrete scalar types
# don't actually tag any values with this
class ScalarT(ConcreteT):
  rank = 0
  _members = ['dtype']

  # no need for _fields_ since scalars aren't represented as a struct
  """
  _properties_ = [
    ('real', identity),
    ('imag', always_zero),
    ('conjugate', identity)
  ]
  """

  def node_init(self):
    assert isinstance(self.dtype, np.dtype), \
        "Expected dtype, got %s" % self.dtype
    self.name = self.dtype.name
    self.nbytes = self.dtype.itemsize

  @property
  def ctypes_repr(self):
    return dtypes.to_ctypes(self.dtype)

  def __eq__(self, other):
    return other.__class__ is self.__class__ and self.nbytes == other.nbytes
    
  def __hash__(self):
    return hash(self.dtype)

  def __repr__(self):
    return self.name

  def __str__(self):
    return str(self.name)

  def index_type(self, _):
    return self

  def combine(self, other):
    if isinstance(other, ScalarT):
      combined_dtype = np.promote_types(self.dtype, other.dtype)
      if combined_dtype == self.dtype:
        return self
      elif combined_dtype == other.dtype:
        return other
      else:
        return from_dtype(combined_dtype)
    else:
      raise IncompatibleTypes(self, other)

_dtype_to_parakeet_type = {}
def register_scalar_type(ParakeetClass, dtype, equiv_python_types = []):
  parakeet_type = ParakeetClass(dtype)
  _dtype_to_parakeet_type[dtype] = parakeet_type

  python_types = [dtype.type] + equiv_python_types

  for python_type in python_types:
    type_conv.register(python_type, parakeet_type)

  return parakeet_type



class IntT(ScalarT):
  """Base class for bool, signed and unsigned"""

  _members = []

class BoolT(IntT):
  """
  The type is called BoolT to distinguish it from its only instantiation called
  Bool.
  """

  def node_init(self):
    assert dtypes.is_bool(self.dtype)
    self.name = 'bool'

  def __eq__(self, other):
    return other.__class__ is BoolT

Bool = register_scalar_type(BoolT, dtypes.bool8, equiv_python_types = [bool])

class UnsignedT(IntT):
  def node_init(self):
    assert dtypes.is_unsigned(self.dtype), \
        "Expected unsigned but got %s" % self.dtype

UInt8 = register_scalar_type(UnsignedT, dtypes.uint8)

UInt16 = register_scalar_type(UnsignedT, dtypes.uint16)
UInt32 = register_scalar_type(UnsignedT, dtypes.uint32)
UInt64 = register_scalar_type(UnsignedT, dtypes.uint64)

class SignedT(IntT):
  def node_init(self):
    assert dtypes.is_signed(self.dtype)

Int8 = register_scalar_type(SignedT, dtypes.int8)
Int16 = register_scalar_type(SignedT, dtypes.int16)
Int32 = register_scalar_type(SignedT, dtypes.int32)
Int64 = register_scalar_type(SignedT, dtypes.int64,
                             equiv_python_types = [int, long])

class FloatT(ScalarT):
  _members = []
  def node_init(self):
    assert dtypes.is_float(self.dtype)

Float32 = register_scalar_type(FloatT, dtypes.float32)
Float64 = register_scalar_type(FloatT, dtypes.float64,
                               equiv_python_types = [float])

def complex_conj(x):
  np.complex(x.real, -x.imag)

class ComplexT(ScalarT):
  _members = ['elt_type']
  _props_ = [('conjugate', complex_conj)]

  def node_init(self):
    assert dtypes.is_float(self.elt_type), \
        "Expected fields of complex to be floating, got %s" % self.elt_type
    assert dtypes.is_complex(self.dtype)
    self._fields_ = [('real', self.elt_type), ('imag', self.elt_type)]

Complex64 = ComplexT(dtypes.float32, dtypes.complex64)
Complex128 = ComplexT(dtypes.float64, dtypes.complex128)

class ConstIntT(IntT):
  """
  Integer constants get a special type all to their lonesome selves, so we can
  assign types to expressions like "(1,2,3)[0]".
  """

  _members = ['value']

  def __init__(self, value):
    self.dtype = dtypes.int64
    self.value = value

  def combine(self, other):
    if isinstance(other, ConstIntT) and other.value == self.value:
      return self
    else:
      return IntT.combine(self, other)

def is_scalar_subtype(t1, t2):
  return isinstance(t1, ScalarT) and \
         isinstance(t2, ScalarT) and \
         ((t1 == t2) or (t1.nbytes() < t2.nbytes()) or \
          (isinstance(t1, IntT) and isinstance(t2, FloatT)))

def register_numeric_type(klass, dtype):
  parakeet_type = klass(dtype)
  _dtype_to_parakeet_type[dtype] = parakeet_type
  return parakeet_type

def from_dtype(dtype):
  return _dtype_to_parakeet_type[dtype]

def from_char_code(c):
  numpy_type = np.typeDict[c]
  return from_dtype(np.dtype(numpy_type))

def is_scalar(t):
  return isinstance(t, ScalarT)

def all_scalars(ts):
  return all(map(is_scalar, ts))

###########################################
#
#  Pointers!
#
###########################################

class PtrT(ConcreteT):
  """
  I'm not giving pointer a concrete to_python/from_python conversion or any
  usable fields, so it's up to our ctypes_repr and llvm_backend to appropriately
  interpret objects of this type.
  """

  _members = ['elt_type']

  rank = 1
  def index_type(self, idx):
    assert isinstance(idx, IntT), \
        "Index into pointer must be of type int, got %s" % (idx)
    return self.elt_type

  def node_init(self):
    self._ctypes_repr = ctypes.POINTER(self.elt_type.ctypes_repr)

  def __str__(self):
    return "ptr(%s)" % self.elt_type

  def __eq__(self, other):
    return isinstance(other, PtrT) and self.elt_type == other.elt_type

  def __hash__(self):
    return hash(self.elt_type)

  def __repr__(self):
    return str(self)

  @property
  def ctypes_repr(self):
    return self._ctypes_repr

_ptr_types = {}
def ptr_type(t):
  if t in _ptr_types:
    return _ptr_types[t]
  else:
    ptr_t = PtrT(t)
    _ptr_types[t] = ptr_t
    return ptr_t



from array_type import ArrayT 
from core_types import PtrT
from syntax import Var, Alloc, ArrayView, Array, Struct, AllocArray  
from syntax import Adverb
from syntax_visitor import SyntaxVisitor

class FindLocalArrays(SyntaxVisitor):
  def __init__(self):
    # hash table mapping from variable names to
    # statements allocating space
    self.local_allocs = {}

    # hash table mapping from variable names to
    # places where we create array views containing
    # locally allocated data pointers
    self.local_arrays = {}
    
    # associate each local array 
    # with its data allocation 
    self.array_to_alloc = {}
  
  def visit_merge(self, merge):
    """
    If both sides of a flow merge are a locally created 
    array originating on the same statement, then 
    also mark the new phi-bound variable as also being 
    that local array. 
    
    Similarly, if both sides are the same local pointer 
    allocation, then propagate that information to the 
    new variable being created by a phi-node
    """
    for (new_name, (left, right)) in merge.iteritems():
      if left.__class__ is Var and right.__class__ is Var:
        left_name = left.name 
        right_name = right.name 
        if left.type.__class__ is ArrayT and \
           left_name in self.local_arrays and \
           right_name in self.local_arrays:
          left_stmt = self.local_arrays[left_name]
          right_stmt = self.local_arrays[right_name]
          if left_stmt == right_stmt:
            self.local_arrays[new_name] = left_stmt 
            if left.name in self.array_to_alloc_name:
              self.array_to_alloc[new_name] = \
                  self.array_to_alloc[left.name]
        elif left.type.__class__ is PtrT and \
             left_name in self.local_allocs and \
             right_name in self.local_allocs:
          left_stmt = self.local_allocs[left_name]
          right_stmt = self.local_allocs[right_name]
          if left_stmt == right_stmt:
            self.local_allocs[new_name] = left_stmt      
  
  def visit_Assign(self, stmt):
    if stmt.lhs.__class__ is Var:
      lhs_name = stmt.lhs.name 
      rhs_class = stmt.rhs.__class__
      if rhs_class is Alloc:
        self.local_allocs[lhs_name] = stmt
      elif rhs_class is ArrayView and \
          stmt.rhs.data.__class__ is Var:
        data_name = stmt.rhs.data.name 
        if data_name in self.local_allocs:
          self.local_arrays[lhs_name] = stmt
          self.array_to_alloc[lhs_name] = data_name  
      elif rhs_class is Struct and \
          stmt.rhs.type.__class__ is ArrayT and \
          stmt.rhs.args[0].__class__ is Var:
        data_name = stmt.rhs.args[0].name
        if data_name in self.local_allocs:
          self.local_arrays[lhs_name] = stmt
          self.array_to_alloc[lhs_name] = data_name 
      elif rhs_class is Array or AllocArray:
        self.local_arrays[stmt.lhs.name] = stmt
      elif isinstance(stmt.rhs, Adverb):
        self.local_arrays[lhs_name] = stmt 


"""
Generic traversal of a tree-structured type
"""

class Traversal(object):
  def visit_list(self, xs):
    return [self.visit(x) for x in xs]
  
  def visit_tuple(self, xs):
    return tuple(self.visit_list(xs))
  
  def visit_generic(self, x):
    assert False, \
      "Unsupported %s : %s" % (x, x.__class__.__name__)
  
  def visit(self, x):
    method_name = 'visit_' + x.__class__.__name__
    
    if hasattr(self, method_name):
      method = getattr(self, method_name)
      return method(x)
    else:
      return self.visit_generic(x)
      


import core_types


from args import ActualArgs 
from node import Node 

class Expr(Node):
  _members = ['type']

  
  def children(self):
    for v in self.itervalues():
      if v and isinstance(v, Expr):
        yield v
      elif isinstance(v, (list,tuple)):
        for child in v:
          if isinstance(child, Expr):
            yield child

  def short_str(self):
    return str(self)

class Const(Expr):
  _members = ['value']
  
  def children(self):
    return (self.value,)

  def short_str(self):
    return str(self.value)

  def __repr__(self):
    if self.type and not isinstance(self.type, core_types.NoneT):
      return "%s : %s" % (self.value, self.type)
    else:
      return str(self.value)

  def __str__(self):
    return repr(self)

  def __hash__(self):
    return hash(self.value)

  def __eq__(self, other):
    return other.__class__ is Const and \
           self.value == other.value and \
           self.type == other.type
    
  def __ne__(self, other):
    return other.__class__ is not Const or \
           self.value != other.value or \
           self.type != other.type

class Var(Expr):
  _members = ['name']

  def node_init(self):
    assert self.name is not None
    
  def short_str(self):
    return self.name

  def __repr__(self):
    if hasattr(self, 'type'):
      return "%s : %s" % (self.name, self.type)
    else:
      return "%s" % self.name

  def __str__(self):
    return self.name

  def __hash__(self):
    return hash(self.name)

  def __eq__(self, other):
    return other.__class__  is Var and \
           self.name == other.name and \
           self.type == other.type


  def __ne__(self, other):
    return other.__class__ is not Var or \
           self.name != other.name or \
           self.type != other.type
  
  def children(self):
    return ()

class Attribute(Expr):
  _members = ['value', 'name']

  def children(self):
    yield self.value

  def __str__(self):
    return "attr(%s, '%s')" % (self.value, self.name)

  def __hash__(self):
    return hash ((self.value, self.name))

  def __eq__(self, other):
    return other.__class__ is Attribute and \
           self.name == other.name and \
           self.value == other.value

class Index(Expr):
  _members = ['value', 'index']
  
  def __eq__(self, other):
    return other.__class__ is Index and \
           other.value == self.value and \
           other.index == self.index
  
  def __hash__(self):
    return hash((self.value, self.index))
  
  def children(self):
    yield self.value
    yield self.index

  def __str__(self):
    return "%s[%s]" % (self.value, self.index)

class Tuple(Expr):
  _members = ['elts']

  def node_init(self):
    self.elts = tuple(self.elts)
    if self.type is None and all(e.type is not None for e in self.elts):
      import tuple_type 
      self.type = tuple_type.make_tuple_type(tuple(e.type for e in self.elts))
      
  def __str__(self):
    if len(self.elts) > 0:
      return ", ".join([str(e) for e in self.elts])
    else:
      return "()"
    
  def children(self):
    return self.elts

  def __hash__(self):
    return hash(self.elts)

class Array(Expr):
  _members = ['elts']

  def node_init(self):
    self.elts = tuple(self.elts)

  def children(self):
    return self.elts

  def __hash__(self):
    return hash(self.elts)

class Closure(Expr):
  """Create a closure which points to a global fn with a list of partial args"""

  _members = ['fn', 'args']

  def __str__(self):
    fn_str = str(self.fn) #self.fn.name if hasattr(self.fn, 'name') else str(self.fn)
    args_str = ",".join(str(arg) + ":" + str(arg.type) for arg in self.args)
    return "Closure(%s, fixed_args = {%s})" % (fn_str, args_str)

  def node_init(self):
    self.args = tuple(self.args)

  def children(self):
    if isinstance(self.fn, Expr):
      yield self.fn
    for arg in self.args:
      yield arg

  def __hash__(self):
    return hash((self.fn, tuple(self.args)))

class Call(Expr):
  _members = ['fn', 'args']

  def __str__(self):
    #if isinstance(self.fn, (Fn, TypedFn)):
    #  fn_name = self.fn.name
    #else:
    fn_name = str(self.fn)
    if isinstance(self.args, ActualArgs):
      arg_str = str(self.args)
    else:
      arg_str = ", ".join(str(arg) for arg in self.args)
    return "%s(%s)" % (fn_name, arg_str)

  def __repr__(self):
    return str(self)

  def children(self):
    yield self.fn
    for arg in self.args:
      yield arg

  def __hash__(self):
    return hash((self.fn, tuple(self.args)))

class Slice(Expr):
  _members = ['start', 'stop', 'step']

  def __str__(self):
    return "slice(%s, %s, %s)"  % \
        (self.start.short_str(),
         self.stop.short_str(),
         self.step.short_str())

  def __repr__(self):
    return str(self)

  def children(self):
    yield self.start
    yield self.stop
    yield self.step

  def __eq__(self, other):
    return other.__class__ is Slice and \
           other.start == self.start and \
           other.stop == self.stop and \
           other.step == self.step
  def __hash__(self):
    return hash((self.start, self.stop, self.step))

class PrimCall(Expr):
  """
  Call a primitive function, the "prim" field should be a prims.Prim object
  """
  _members = ['prim', 'args']
  
  def _arg_str(self, i):
    arg = self.args[i]
    if arg.__class__ is PrimCall:
      return "(%s)" % arg
    else:
      return str(arg)
  
  def __hash__(self):
    self.args = tuple(self.args)
    return hash((self.prim, self.args))
    
  def __eq__(self, other):
    if other.__class__ is not PrimCall:
      return False 
    
    if other.prim != self.prim:
      return False
    
    my_args = self.args
    other_args = other.args
    n = len(my_args)
    if n != len(other_args):
      return False
    
    for (i,arg) in enumerate(my_args):
      other_arg = other_args[i]
      if arg != other_arg:
        return False
    return True
  
  def __ne__(self, other):
    return not self==other
  
  def __repr__(self):
    if self.prim.symbol:
      if len(self.args) == 1:
        return "%s %s" % (self._arg_str(0), self.prim.symbol)
      else:
        assert len(self.args) == 2
        return "%s %s %s" % (self._arg_str(0), self.prim.symbol, self._arg_str(1))
    else:
      arg_strings = [self._arg_str(i) for i in xrange(len(self.args))]
      combined = ", ".join(arg_strings)
      return "prim<%s>(%s)" % (self.prim.name, combined)

  def __str__(self):
    return repr(self)

  def node_init(self):
    self.args = tuple(self.args)

  def children(self):
    return self.args
  


class TupleProj(Expr):
  _members = ['tuple', 'index']

  def node_init(self):
    if self.type is None:
      if self.tuple.type is not None and self.index.__class__ is Const:
        self.type = self.tuple.type.elt_types[self.index.value]
            
  def __str__(self):
    return "TupleProj(%s, %d)" % (self.tuple, self.index)

  def children(self):
    return (self.tuple,)

  def __hash__(self):
    return hash((self.tuple, self.index))

class ClosureElt(Expr):
  _members = ['closure', 'index']

  def __str__(self):
    return "ClosureElt(%s, %d)" % (self.closure, self.index)

  def children(self):
    return (self.closure,)

  def __hash__(self):
    return hash((self.closure, self.index))

class Cast(Expr):
  # inherits the member 'type' from Expr, but for Cast nodes it is mandatory
  _members = ['value']

  def __hash__(self):
    return hash(self.value)
  
  def __str__(self):
    return "Cast(%s : %s)" % (self.value, self.type) 

class Struct(Expr):
  """
  Eventually all non-scalar data should be transformed to be created with this
  syntax node, signifying explicit struct allocation
  """

  _members = ['args']

  def __str__(self):
    return "Struct(%s) : %s" % \
           (", ".join(str(arg) for arg in self.args), self.type)

  def children(self):
    return self.args

  def __hash__(self):
    self.args = tuple(self.args)
    return hash(self.args)

class Alloc(Expr):
  """Allocates a block of data, returns a pointer"""

  _members = ['elt_type', 'count']

  def __str__(self):
    return "alloc<%s>[%s] : %s" % (self.elt_type, self.count, self.type)

  def children(self):
    return (self.count,)

  def __hash__(self):
    return hash((self.elt_type, self.count))


import numpy as np
import scipy.spatial 
import time 

import parakeet 
import testing_helpers

def sqr_dist(x,y):
  return sum( (x-y) ** 2)

def apd(X,Y):
  return parakeet.allpairs(sqr_dist, X, Y)



def test_dists():
  m = 50 
  n = 20
  d = 10
  X = np.random.randn(m,d)
  Y = np.random.randn(n,d)

  start = time.time()
  _ = apd(X,Y)
  parakeet_time = time.time() - start 
  
  start = time.time()
  parakeet_dists = apd(X,Y)
  parakeet_no_comp_time = time.time() - start
  
  start = time.time()
  python_dists = scipy.spatial.distance.cdist(X,Y,'sqeuclidean')
  np_time = time.time() - start 
  
  print "Parakeet time", parakeet_time
  print "Parakeet (no compilation)", parakeet_no_comp_time
  print "NumPy time", np_time
  
  assert testing_helpers.eq(parakeet_dists, python_dists)    
  
if __name__ == '__main__':
  testing_helpers.run_local_tests()



import numpy as np
import time 

import parakeet 

from testing_helpers import expect_each, run_local_tests, eq

size = (5,5)
float_mat = np.random.uniform(0,1,size=size)
bool_mat = float_mat > 0.5
int_mat = np.random.random_integers(0,255,size=size)

matrices = [float_mat, bool_mat, int_mat]

def diff_x(I):
  m = I.shape[0]
  return (I[1:, :] - I[:m-1, :])[:, 1:]

def test_diff_x():
  expect_each(diff_x, diff_x, matrices)

def diff_y(I):
  n = I.shape[1]
  return (I[:, 1:] - I[:, :n-1])[1:, :]

def test_diff_y():
  expect_each(diff_x, diff_x, matrices)


def harris(I):
  dx = diff_x(I)
  dy = diff_y(I)
  #
  #   At each point we build a matrix
  #   of derivative products
  #   M =
  #   | A = dx^2     C = dx * dy |
  #   | C = dy * dx  B = dy * dy |
  #
  #   and the score at that point is:
  #      det(M) - k*trace(M)^2
  #
  A = dx * dx
  B = dy * dy
  C = dx * dy
  tr = A + B
  det = A * B - C * C
  k = 0.05
  return det - k * tr * tr

def parallel_harris(I):
   
  dx = diff_x(I)
  dy = diff_y(I)
  def scalar_computation(dx,dy):
    A = dx * dx
    B = dy * dy
    C = dx * dy
    tr = A + B
    det = A * B - C * C
    k = 0.05
    return det - k * tr * tr
  return parakeet.each(scalar_computation, dx, dy)


def test_harris():
  expect_each(harris, harris, matrices)

def test_harris_timing():
  x = np.random.randn(1500, 1500)
  
  np_start = time.time()
  harris(x)
  np_time = time.time() - np_start

  seq_start = time.time()
  parakeet.run(harris,x)
  seq_time = time.time() - seq_start

  seq_start_no_comp = time.time()
  parakeet.run(harris,x)
  seq_time_no_comp = time.time() - seq_start_no_comp

  par_start = time.time()
  parallel_harris(x)
  par_time = time.time() - par_start
  
  par_start_no_comp = time.time()
  parallel_harris(x)

  par_time_no_comp = time.time() - par_start_no_comp
  
  print "Parakeet time: %.3f" % seq_time
  print "Parakeet w/out compilation: %.3f" % seq_time_no_comp
  print "Parakeet parallel time: %.3f" % par_time
  print "Parakeet parallel time (no comp): %.3f" % par_time_no_comp
  print "Python time: %.3f" % np_time
  assert par_time_no_comp / np_time < 5, \
    "Parakeet too slow (%.1fX slowdown)" % (par_time_no_comp / np_time)
  
  """
  start = time.time()
  parallel_result = parallel_harris(x)
  parallel_time = time.time() - start
  print "Parallel time: %.3f" % parallel_time
  rmse = np.sqrt(np.mean( (parallel_result - np_rslt) **2))
  assert rmse < 0.001, rmse 
  """
  
if __name__ == '__main__':
  run_local_tests()



from syntax import Fn

# python value of a user-defined function mapped to its
# untyped representation
known_python_functions = {}

def register_python_fn(fn_val, fundef):
  known_python_functions[fn_val] = fundef

def already_registered_python_fn(fn_val):
  return fn_val in known_python_functions

def lookup_python_fn(fn_val):
  """Returns untyped function definition"""

  return known_python_functions[fn_val]



import core_types
import names
import numpy as np
import syntax

from args import FormalArgs

prim_lookup_by_value = {}

def find_prim_from_python_value(fn):
  return prim_lookup_by_value[fn]

prim_lookup_by_op_name = {}

def op_name(op):
  return op.__class__.__name__

def is_ast_op(op):
  return op_name(op) in prim_lookup_by_op_name

def find_ast_op(op):
  name = op_name(op)
  if name in prim_lookup_by_op_name:
    return prim_lookup_by_op_name[name]
  else:
    raise RuntimeError("Operator not implemented: %s" % name)

def is_prim(numpy_fn):
  return numpy_fn in prim_lookup_by_value

_untyped_prim_wrappers = {}
def prim_wrapper(p):
  """Given a primitive, return an untyped function which calls that prim"""
  if p in _untyped_prim_wrappers:
    return _untyped_prim_wrappers[p]
  else:
    fn_name = names.fresh(p.name)

    args_obj = FormalArgs()

    arg_vars = []
    for name in names.fresh_list(p.nin):
      args_obj.add_positional(name)
      arg_vars.append(syntax.Var(name))
    body = [syntax.Return(syntax.PrimCall(p, arg_vars))]
    fundef = syntax.Fn(fn_name, args_obj, body, [])
    _untyped_prim_wrappers[p] = fundef

    return fundef

class Prim(object):
  def __init__(self, fn, python_op_name = None, symbol = None,
               name = None, nin = None, nout = None, 
               extra_signatures = []):
    self.fn = fn
    prim_lookup_by_value[fn] = self
    self.symbol = symbol
    self.python_op_name = python_op_name

    if python_op_name is not None:
      prim_lookup_by_op_name[python_op_name] = self

    if name:
      self.name = name
    else:
      self.name = fn.__name__

    if nin:
      self.nin = nin
    elif hasattr(fn, 'nin'):
      self.nin = fn.nin
    else:
      assert hasattr(fn, 'func_code')
      self.nin = fn.func_code.co_argcount

    if nout:
      self.nout = nout
    elif hasattr(fn, 'nout'):
      self.nout = fn.nout
    else:
      self.nout = 1

    self._create_type_table()
    for sig in extra_signatures:
      self._add_signature(sig)
      
    self.wrapper = prim_wrapper(self)

  def _add_signature(self, signature):
    # numpy type signatures look like 'ff->f' where each character
    # represents a single type

    arg_codes, result_code = signature.split('->')
    try:
      parakeet_types = [core_types.from_char_code(c) for c in arg_codes]
      input_types = tuple(parakeet_types)
      result_type = core_types.from_char_code(result_code)
      self.type_table[input_types] = result_type
    except:
      # print "Signature %s failed  for %s" % (signature , self.fn)
      pass
  
  
  def _create_type_table(self):
    # for now only support ufuncs which describe their own type behavior
    if hasattr(self.fn, 'types'):
      "Primitive function %s doesn't supply type signatures" % self.name

      self.type_table = {}
      for signature in self.fn.types:
        self._add_signature(signature)

  def __eq__(self, other):
    return self.fn == other.fn

  def __hash__(self):
    return hash(self.name)

  def __call__(self, *args, **kwds):
    return self.fn(*args, **kwds)

  def __repr__(self):
    return "prim(%s)" % self.name

  def expected_input_types(self, arg_types):
    """Given some argument types, return the desired upcast types"""
    # by default we just figure out the common type and expect every arg to be
    # of that type
    n_inputs = len(arg_types)
    assert n_inputs == self.nin, \
        "Incorrect number of argument types for %s, expected %s but given %d" \
        % (self.name, self.nin, n_inputs)
    common_type = core_types.combine_type_list(arg_types)
    return [common_type] * n_inputs

  def result_type(self, arg_types):
    """
    Given some argument types, look up the result type in the type_table we
    generated from numpy's given signatures
    """
    key = tuple(arg_types)
    if key not in self.type_table:
      raise RuntimeError("Primitives %s doesn't support input types %s, candidates: %s" % (self.name, key, self.type_table))
    else:
      return self.type_table[key]

class Float(Prim):
  """Always returns a float"""
  pass

class Arith(Prim):
  """Basic arithmetic operators"""
  pass

class Logical(Prim):
  """Expects boolean inputs, returns a boolean"""
  def expected_input_types(self, arg_types):
    return [core_types.Bool] * len(arg_types)

class Bitwise(Prim):
  """Takes any two identical scalar types, returns the same"""
  pass

class Cmp(Prim):
  """Takes two arguments of any type, returns a boolean"""
  pass

class_list = [Cmp, Bitwise, Logical, Arith, Float]

abs = Float(np.abs)
sqrt = Float(np.sqrt)
exp = Float(np.exp)
log = Float(np.log)
sqrt = Float(np.sqrt)
log10 = Float(np.log10)
log2 = Float(np.log2)
cos = Float(np.cos)
cosh = Float(np.cosh)
sin = Float(np.sin)
sinh = Float(np.sinh)
# TODO: figure out how to derive type table for this:
# sinc = Float(np.sinc)
tan = Float(np.tan)
tanh = Float(np.tanh)


logical_and = Logical(np.logical_and, "And")
logical_not = Logical(np.logical_not, "Not")
logical_or = Logical(np.logical_or, "Or")
#logical_xor = Logical(np.logical_xor, 'BitXor')

bitwise_not = Bitwise(np.bitwise_not, 'Invert', '!')
bitwise_and = Bitwise(np.bitwise_and, 'BitAnd', '&')
bitwise_or = Bitwise(np.bitwise_or, 'BitOr', '|')
bitwise_xor = Bitwise(np.bitwise_xor, 'BitXor', '^')

add = Arith(np.add, 'Add', '+')
subtract = Arith(np.subtract, 'Sub', '-')
multiply = Arith(np.multiply, 'Mult', '*')
divide = Arith(np.divide, 'Div', '/', extra_signatures = ['??->?'])


mod = Arith(np.mod, 'Mod', '%', extra_signatures = ['??->?'])


power = Arith(np.power, 'Pow', '**')
negative = Arith(np.negative, 'USub', '-', None, 1, 1)
maximum = Arith(np.maximum, None, None, 'maximum', 2, 1)
minimum = Arith(np.minimum, None, None, 'minimum', 2, 1)


equal = Cmp(np.equal, 'Eq', '==')
not_equal = Cmp(np.not_equal, 'NotEq', '!=')
less = Cmp(np.less, 'Lt', '<')
less_equal = Cmp(np.less_equal, 'LtE', '<=')
greater = Cmp(np.greater, 'Gt', '>')
greater_equal = Cmp(np.greater_equal, 'GtE', '>=')

is_ = Cmp(lambda x,y: x is y, 'Is', 'is')





import inline
import names
import syntax_helpers

from syntax import Var, Const,  Return, TypedFn, DataAdverb, Adverb
from syntax import IndexMap, IndexReduce, Map, Reduce, AllPairs
from transform import Transform
from use_analysis import use_count

def fuse(prev_fn, prev_fixed_args, next_fn, next_fixed_args, fusion_args):
  if syntax_helpers.is_identity_fn(next_fn):
    assert len(next_fixed_args) == 0
    return prev_fn, prev_fixed_args

  """
  Expects the prev_fn's returned value to be one or more of the arguments to
  next_fn. Any element in 'const_args' which is None gets replaced by the
  returned Var
  """
  fused_formals = []
  fused_input_types = []
  fused_type_env = prev_fn.type_env.copy()
  fused_name = names.fresh('fused')

  prev_closure_formals = prev_fn.arg_names[:len(prev_fixed_args)]

  for prev_closure_arg_name in prev_closure_formals:
    t = prev_fn.type_env[prev_closure_arg_name]
    fused_formals.append(prev_closure_arg_name)
    fused_input_types.append(t)

  next_closure_formals = next_fn.arg_names[:len(next_fixed_args)]
  for next_closure_arg_name in next_closure_formals:
    t = next_fn.type_env[next_closure_arg_name]
    new_name = names.refresh(next_closure_arg_name)
    fused_type_env[new_name] = t
    fused_formals.append(new_name)
    fused_input_types.append(t)

  prev_direct_formals = prev_fn.arg_names[len(prev_fixed_args):]
  for arg_name in prev_direct_formals:
    t = prev_fn.type_env[arg_name]
    fused_formals.append(arg_name)
    fused_input_types.append(t)

  prev_return_var, fused_body = \
      inline.replace_return_with_var(prev_fn.body,
                                     fused_type_env,
                                     prev_fn.return_type)
  # for now we're restricting both functions to have a single return at the
  # outermost scope
  inline_args = list(next_closure_formals)
  for arg in fusion_args:
    if arg is None:
      inline_args.append(prev_return_var)
    elif isinstance(arg, int):
      # positional arg which is not being fused out
      inner_name = next_fn.arg_names[arg]

      inner_type = next_fn.type_env[inner_name]
      new_name = names.refresh(inner_name)
      fused_formals.append(new_name)
      fused_type_env[new_name] = inner_type
      fused_input_types.append(inner_type)
      var = Var(new_name, inner_type)
      inline_args.append(var)
    else:
      assert arg.__class__ is Const, \
         "Only scalars can be spliced as literals into a fused fn: %s" % arg
      inline_args.append(arg)
  next_return_var = inline.do_inline(next_fn, inline_args,
                                     fused_type_env, fused_body)
  fused_body.append(Return(next_return_var))

  # we're not renaming variables that originate from the predecessor function
  new_fn = TypedFn(name = fused_name,
                   arg_names = fused_formals,
                   body = fused_body,
                   input_types = tuple(fused_input_types),
                   return_type = next_fn.return_type,
                   type_env = fused_type_env)

  combined_args = prev_fixed_args + next_fixed_args
  return new_fn, combined_args 

class Fusion(Transform):
  def __init__(self, recursive=True):
    Transform.__init__(self)
    # name of variable -> Map or Scan adverb
    self.adverb_bindings = {}
    self.recursive = True

  def pre_apply(self, fn):
    # map each variable to
    self.use_counts = use_count(fn)

  def transform_TypedFn(self, fn):
    if self.fn.copied_by is not None:
      return self.fn.copied_by.apply(fn)
    else:

      # at the very least do high level optimizations
      import pipeline
      return pipeline.high_level_optimizations(fn)

  def transform_Assign(self, stmt):
    if self.recursive:
      stmt.rhs = self.transform_expr(stmt.rhs)
    rhs = stmt.rhs
    if isinstance(rhs, DataAdverb) and \
       rhs.__class__ is not AllPairs and \
       inline.can_inline(self.get_fn(rhs.fn)):

      args = rhs.args
      if all(arg.__class__ in (Var, Const) for arg in args):
        arg_names = [arg.name for arg in args if arg.__class__ is Var]
        unique_vars = set(arg_names)
        adverb_vars = [name for name in unique_vars
                       if name in self.adverb_bindings]

        for arg_name in adverb_vars:
          n_occurrences = sum((name == arg_name for name in arg_names))
          if self.use_counts[arg_name] == n_occurrences:
            prev_adverb = self.adverb_bindings[arg_name]
            if inline.can_inline(self.get_fn(prev_adverb.fn)):
             
              # 
              # Map(Map) -> Map
              # Reduce(Map) -> Reduce 
              # Scan(Map) -> Scan 
              # 
              if prev_adverb.__class__ is Map and rhs.axis == prev_adverb.axis:
                surviving_array_args = []
                fusion_args = []
                for (pos, arg) in enumerate(rhs.args):
                  c = arg.__class__
                  if c is Var and arg.name == arg_name:
                    fusion_args.append(None)
                  elif c is Const:
                    fusion_args.append(arg)
                  else:
                    surviving_array_args.append(arg)
                    fusion_args.append(pos)
                new_fn, clos_args = \
                  fuse(self.get_fn(prev_adverb.fn),
                        self.closure_elts(prev_adverb.fn),
                        self.get_fn(rhs.fn),
                        self.closure_elts(rhs.fn),
                        fusion_args)
                assert new_fn.return_type == self.return_type(rhs.fn)
                del self.adverb_bindings[arg_name]
                if self.fn.copied_by:
                  new_fn = self.fn.copied_by.apply(new_fn)
                rhs.fn = self.closure(new_fn, clos_args)
                rhs.args = prev_adverb.args + surviving_array_args
                
              # 
              # Reduce(IndexMap) -> IndexReduce
              #   
              elif prev_adverb.__class__ is IndexMap and \
                   rhs.__class__ is Reduce and \
                   len(rhs.args) == 1 and \
                   (self.is_none(rhs.axis) or rhs.args[0].type.rank == 1):
                
                new_fn, clos_args = \
                  fuse(self.get_fn(prev_adverb.fn),
                        self.closure_elts(prev_adverb.fn),
                        self.get_fn(rhs.fn),
                        self.closure_elts(rhs.fn),
                        rhs.args)
                assert new_fn.return_type == self.return_type(rhs.fn)
                del self.adverb_bindings[arg_name]
                if self.fn.copied_by:
                  new_fn = self.fn.copied_by.apply(new_fn)
                stmt.rhs = IndexReduce(fn = self.closure(new_fn, clos_args), 
                                  shape = prev_adverb.shape, 
                                  combine = rhs.combine, 
                                  type = rhs.type,
                                  init = rhs.init )
                
    if stmt.lhs.__class__ is Var and isinstance(rhs, Adverb):
      self.adverb_bindings[stmt.lhs.name] = rhs
    return stmt



import sys
def dispatch(node, prefix = "",  default = None, locals_dict = None):
  if locals_dict is None:
    last_frame = sys._getframe()
    locals_dict = last_frame.f_back.f_locals
  node_type = node.__class__.__name__
  if len(prefix) > 0:
    fn_name = prefix + "_" + node_type
  else:
    fn_name = node_type

  if fn_name in locals_dict:
    return locals_dict[fn_name]()
  elif default:
    return default(node)
  else:
    available = [k.split(prefix + "_")[1] for k in locals_dict.keys() if k.startswith(prefix + "_")]
    raise RuntimeError("Unsupported node %s with type %s, available: %s" % (node, node_type, available))

import ctypes
def list_to_ctypes_array(ctypes_object_list, elt_type = None, pointers = False):
  if elt_type is None:
    elt = ctypes_object_list[0]
    elt_type = elt.__class__
  if pointers:
    elt_type = ctypes.POINTER(elt_type)
  array_t = elt_type * len(ctypes_object_list)
  array_object = array_t()
  for (i, x) in enumerate(ctypes_object_list):
    array_object[i] = ctypes.pointer(x) if pointers else x
  return array_object



import config
import syntax 
import syntax_visitor

from copy_elimination import CopyElimination
from dead_code_elim import DCE
from fusion import Fusion
from inline import Inliner
from licm import LoopInvariantCodeMotion
from loop_unrolling import LoopUnrolling
from lower_adverbs import LowerAdverbs
from lower_indexing import LowerIndexing
from lower_structs import LowerStructs
from mapify_allpairs import MapifyAllPairs
from pipeline_phase import Phase
from value_range_propagation import RangePropagation
from redundant_load_elim import RedundantLoadElimination
from scalar_replacement import ScalarReplacement
from shape_elim import ShapeElimination
from simplify import Simplify
from index_elimination import IndexElim

class ContainsAdverbs(syntax_visitor.SyntaxVisitor):
  class Yes(Exception):
    pass
  
  def visit_expr(self, expr):
    if isinstance(expr, syntax.Adverb):
      raise self.Yes()
    syntax_visitor.SyntaxVisitor.visit_expr(self, expr)
    
def contains_adverbs(fn):
  try:
    ContainsAdverbs().visit_fn(fn)
  except ContainsAdverbs.Yes:
    return True
  return False


fusion_opt = Phase(Fusion, config_param = 'opt_fusion', cleanup = [DCE],
                   memoize = False,
                   run_if = contains_adverbs)
inline_opt = Phase(Inliner, config_param = 'opt_inline', cleanup = [])
high_level_optimizations = Phase([Simplify, inline_opt, Simplify, DCE,
                                  fusion_opt, fusion_opt])
def print_loopy(fn):
  if config.print_loopy_function:
    print
    print "=== Loopy function ==="
    print
    print repr(fn)
    print

copy_elim = Phase(CopyElimination, config_param = 'opt_copy_elimination')
licm = Phase(LoopInvariantCodeMotion, config_param = 'opt_licm',
             memoize = False)

symbolic_range_propagation = Phase(RangePropagation,
                           config_param = 'opt_range_propagation',
                           memoize = False)
shape_elim = Phase(ShapeElimination,
                   config_param = 'opt_shape_elim')
# loop_fusion = Phase(LoopFusion, config_param = 'opt_loop_fusion')

loopify = Phase([Simplify,
                 fusion_opt,
                 LowerAdverbs, inline_opt,
                 copy_elim,
                 licm,],
                depends_on = high_level_optimizations,
                cleanup = [Simplify, DCE],
                copy = True,
                run_if = contains_adverbs,
                post_apply = print_loopy)




scalar_repl = Phase(ScalarReplacement, config_param = 'opt_scalar_replacement')
load_elim = Phase(RedundantLoadElimination,
                  config_param = 'opt_redundant_load_elimination')
unroll = Phase(LoopUnrolling, config_param = 'opt_loop_unrolling')

index_elim = Phase(IndexElim, config_param = 'opt_index_elimination')

pre_lowering = Phase([symbolic_range_propagation,
                      shape_elim,
                      symbolic_range_propagation,
                      index_elim ],
                     cleanup = [Simplify, DCE])
post_lowering = Phase([licm,
                       unroll,
                       licm,
                       Simplify,
                       load_elim,
                       scalar_repl,
                       ], cleanup = [Simplify, DCE])

lowering = Phase([pre_lowering,
                  LowerIndexing,
                  licm,
                  LowerStructs,
                  post_lowering,
                  LowerStructs,
                  Simplify],
                 depends_on = loopify,
                 copy = True,
                 cleanup = [])



import parakeet 
import testing_helpers

import numpy as np 

def init_layer(input_size, n_elts):
  return np.random.randn(n_elts, input_size)

def create_mlp(n_inputs, n_hidden, n_outputs):
  W1 = init_layer(n_inputs, n_hidden)
  W2 = init_layer(n_hidden, n_outputs)
  return (W1, W2)

def dot(x,y):
  return sum(x*y)

def fprop_elt(x,w):
  return parakeet.tanh(dot(x,w))

def fprop_layer(layer, x):
  return [fprop_elt(x,w) for w in layer]

def fprop(network, x):
  for layer in network:
    x = parakeet.run(fprop_layer, layer, x)
  return x

def fprop_python(network, x): 
  for layer in network:
    x = np.array(fprop_layer(layer,x))
  return x

n = 1000
d = 500
Xpos = np.random.randn(n,d)+1.0
Xneg = np.random.randn(n,d)-1.0
X = np.vstack([Xpos, Xneg])
Y = np.vstack([np.ones(n), np.ones(n)*-1])

network = create_mlp(d,50,1)

def test_fprop():
  parakeet_result = fprop(network, X[0])
  print "Parakeet result", parakeet_result
  python_result = fprop_python(network, X[0])
  print "Python result", python_result
  assert testing_helpers.eq(parakeet_result, python_result)
    
if __name__ == '__main__':
  testing_helpers.run_local_tests()



import numpy as np

import parakeet 
import testing_helpers 

values = [np.array(0), np.array(1.0), np.array([True]), 
          np.array([1,2,3]), np.array([1.0, 2.0, 3.0]), np.array([True, False]),
          np.array([[1,2,3],[10,20,30]]), np.array([[1.0,2.0,3.0],[10.0,20.0, 30.0]]), 
          np.array([[True], [False]])
         ]

def run(fn, method_name):
  fn = parakeet.jit(fn)
  for v in values:
    method = getattr(v, method_name)
    expected = method()
    result = fn(v)
    assert np.allclose(expected, result), "Expected %s but got %s" % (expected, result)

def test_min():
  def call_min(x):
    return x.min()
  run(call_min, 'min')


def test_max():
  def call_max(x):
    return x.max()
  run(call_max, 'max')


def test_any():
  def call_any(x):
    return x.any()
  run(call_any, 'any')

def test_all():
  def call_all(x):
    return x.all()
  run(call_all, 'all')

def test_copy():
  def call_copy(x):
    return x.copy()
  run(call_copy, 'copy')
  x = np.array([1,2,3])
  y = parakeet.jit(call_copy)(x)
  x[0] = 10
  assert y[0] == 1
  
def test_ravel():
  def call_ravel(x):
    return x.ravel()
  run(call_ravel, 'ravel')

def test_flatten():
  def call_flatten(x):
    return x.flatten()
  run(call_flatten, 'flatten')

if __name__ == '__main__':
  testing_helpers.run_local_tests() 



import numpy as np
import pylab 
import scipy
import scipy.ndimage 
import time 

import parakeet

from parakeet import jit, pmap2 
from testing_helpers import eq, run_local_tests


def winmin(x):
  m,n = x.shape
  v = x[0,0]
  for i in range(m):
    for j in range(n):
      v2 = x[i,j]
      if v2 < v:
        v = v2
  return v

def winmax(x):
  m,n = x.shape
  v = x[0,0]
  for i in range(m):
    for j in range(n):
      v2 = x[i,j]
      if v2 > v:
        v = v2
  return v

def erode(X, window_size = (3,3)):
  return pmap2(winmin, X, window_size)
 
def dilate(X, window_size = (3,3)):
  return pmap2(winmax, X, window_size)


def load_img(path  = 'data/rjp_small.png', gray=True):
  try:
    x = pylab.imread(path)
  except:
    x = pylab.imread('../' + path)
  if len(x.shape) > 2 and gray:
    x =  x[:, :, 2]
  if len(x.shape) > 2 and x.shape[2] == 4:
    x = x[:,:,:3]
  if x.max() > 1: 
    x = x.astype('float') / 257.0
  return x


X = np.array([[0, 0,   0,   0,   0],
              [0, 0.1,  0.2, 0.3, 0],
              [0, 0.3,  0.4, 0.3, 0],
              [0, 0.2,  0.1, 0.5, 0],
              [0, 0,    0,   0,   0]])

# what we would expect after a 3x3 erosion 
X_erode = np.array([[0, 0, 0,   0, 0],
                    [0, 0, 0,   0, 0],
                    [0, 0, 0.1, 0, 0],
                    [0, 0, 0,   0, 0],
                    [0, 0, 0,   0, 0]])

X_dilate = np.array([[0.1, 0.2, 0.3, 0.3, 0.3],
                     [0.3, 0.4, 0.4, 0.4, 0.3],
                     [0.3, 0.4, 0.5, 0.5, 0.5],
                     [0.3, 0.4, 0.5, 0.5, 0.5],
                     [0.2, 0.2, 0.5, 0.5, 0.5]])

def test_erode():
  res = erode(X, (3,3))
  assert res.shape == X_erode.shape, "Expected shape %s but got %s" % (X_erode.shape, res.shape)
  assert (res == X_erode).all(), "Expected %s but got %s" % (X_erode, res)

def test_dilate():
  res = dilate(X, (3,3))
  assert res.shape == X_dilate.shape, "Expected shape %s but got %s" % (X_dilate.shape, res.shape)
  assert (res == X_dilate).all(), "Expected %s but got %s" % (X_dilate, res)



def dilate_1d_naive(x_strip, y_strip, window_size):
  """
  Given a 1-dimensional input and 1-dimensional output, 
  fill output with 1d dilation of input 
  """
  nelts = len(x_strip)
  half = window_size / 2 
  for i in xrange(nelts):
    left_idx = max(i-half,0)
    right_idx = min(i+half+1, nelts)
    currmax = x_strip[left_idx]
    for j in xrange(left_idx+1, right_idx):
      elt = x_strip[j]
      if elt > currmax:
        currmax = elt
    y_strip[i] = currmax 

@jit 
def dilate_decompose(x, window_size): 
  m,n = x.shape
  k,l = window_size
  y = np.empty_like(x)
  z = np.empty_like(x)
  for row_idx in xrange(m):
    dilate_1d_naive(x[row_idx,:], y[row_idx,:], k)
  for col_idx in xrange(n):
    dilate_1d_naive(y[:, col_idx], z[:, col_idx], l)
  return z

def test_dilate_decompose():
  res = dilate_decompose(X, (3,3))
  assert res.shape == X_dilate.shape, "Expected shape %s but got %s" % (X_dilate.shape, res.shape)
  assert (res == X_dilate).all(), "Original \n%s Expected dilation \n%s but got \n%s unequal elts \n%s" % \
    (X, X_dilate, res, X_dilate != res)

  """
  x = load_img(gray=False)
  def filter(img):
    return dilate(img, (10,10))
  r,g,b = x[:,:,0],x[:,:,1],x[:,:,2]
  rd = filter(r)
  
  assert rd.shape == r.shape 
  gd = filter(g)
  bd = filter(b)
  if plot:
    pylab.imshow(x)
    pylab.figure()
    y = np.dstack([rd, gd, bd])
    pylab.imshow(y)
    pylab.figure()
    pylab.imshow((y-x)**3)
    pylab.show()
  assert (r <= rd).all(), np.sum(r > rd)
  assert (g <= gd).all(), np.sum(g > gd)
  assert (b <= bd).all(), np.sum(b > bd)
  """

"""
def test_erode():
  x = load_img(gray=False)
  size = (10,10)
  r,g,b = x[:,:,0],x[:,:,1],x[:,:,2]
  

  def filter(img):
    par_start_compile_t = time.time()
    print "---"
    res_par = erode(img, size)
    par_end_compile_t = time.time()
    print "Parakeet compile time: %0.3f" % (par_end_compile_t - par_start_compile_t)
    
    print "---"
    par_start_t = time.time()
    res_par = erode(img, size)
    par_end_t = time.time()
    print "Parakeet time: %0.3f" % (par_end_t - par_start_t)
    sci_start_t = time.time()
    res_sci = scipy.ndimage.grey_erosion(img, size, mode = 'nearest')
    sci_end_t = time.time()
    print "SciPy time: %0.3f" % (sci_end_t - sci_start_t)
    
    
    if plot:
      pylab.imshow(res_par)
      pylab.figure()
      pylab.imshow(res_sci)
      pylab.figure()
      pylab.imshow(res_sci - res_par)
      pylab.show()
    assert res_par.shape == res_sci.shape
    assert (res_par == res_sci).all(), \
      "# different elements: %d / %d" % ((res_par != res_sci).sum(), res_par.size)
    return res_par
  filter(r)
  filter(g)
  filter(b)
  
"""
def morph_open(x, erode_shape, dilate_shape = None):
  if dilate_shape is None: 
    dilate_shape = erode_shape
  return dilate(erode(x, erode_shape), dilate_shape)


def morph_close(x, dilate_shape, erode_shape = None):
  if erode_shape is None: 
    erode_shape = dilate_shape
  return erode(dilate(x, dilate_shape), erode_shape)


plot_rgb = False 

def test_residual():
  x = load_img(gray=False)
  s1 = (5,20)
  s2 = (17,3)
  def filter(img):
    return erode(dilate(img, (3,3)), (1,1))#  + erode(img, s2), s2) **3
    
  r = filter(x[:,:,0])
  g = filter(x[:,:,1])
  b = filter(x[:,:,2])
  y = np.dstack([r,g,b])
  if plot_rgb:
    pylab.imshow(x)
    pylab.figure()
    pylab.imshow(y)
    pylab.show()

if __name__ == '__main__':
  run_local_tests() 



class ScopedSet(object):
  def __init__(self, initial_set = None):
    self.sets = []
    self.push(initial_set if initial_set is not None else [])

  def push(self, new_set = None):
    if new_set is None:
      new_set = set([])
    elif not isinstance(new_set, set):
      new_set = set(new_set)
    self.sets.append(new_set)
        
  def pop(self):
    return self.sets.pop()
  
  def top(self):
    return self.sets[-1]
  
  def add(self, value):
    self.top().add(value)
  
  def update(self, values):
    self.top().update(values)
      
  def __contains__(self, value):
    return value in self.top()
  
  def __str__(self):
    return "ScopedSet(%s)" % ", ".join(str(s) for s in self.sets)
  